from sys import stdout
from typing import Any, List, Tuple, cast
from unittest.mock import Mock, patch

from zope.interface import implementer

from twisted.python.components import registerAdapter
from twisted.trial import unittest
from twisted.web.iweb import IRequest

from .. import Klein
from .._app import KleinRenderable, KleinRequest
from .._decorators import bindable, modified, originalName
from .._interfaces import IKleinRequest
from .test_resource import MockRequest
from .util import EqualityTestsMixin


@implementer(IRequest)
class DummyRequest:  # type: ignore[misc]  # incomplete interface
    def __init__(self, n: object) -> None:
        self.n = n

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DummyRequest):
            return other.n == self.n
        return NotImplemented


registerAdapter(KleinRequest, DummyRequest, IKleinRequest)


class DummyRequestSelfTest(unittest.TestCase):
    """
    Tests for L{DummyRequest} to make sure test failures are meaningful.
    """

    def test_equality(self) -> None:
        """
        Dummy requests are equal to each other if their value matches; they do
        not equal other types.
        """
        self.assertEqual(DummyRequest(3), DummyRequest(3))
        self.assertNotEqual(DummyRequest(3), DummyRequest(4))
        self.assertNotEqual(DummyRequest(3), 3)


class KleinEqualityTestCase(unittest.TestCase, EqualityTestsMixin):
    """
    Tests for L{Klein}'s implementation of C{==} and C{!=}.
    """

    class _One:
        app = Klein()

        def __eq__(self, other: Any) -> bool:
            return True

        def __ne__(self, other: Any) -> bool:
            return False

        def __hash__(self) -> int:
            return id(self)

    _another = Klein()

    def anInstance(self) -> Klein:
        # This is actually a new Klein instance every time since Klein.__get__
        # creates a new Klein instance for every instance it is retrieved from.
        # The different _One instance, at least, will not cause the Klein
        # instances to be not-equal to each other since an instance of _One is
        # equal to everything.
        return self._One().app

    def anotherInstance(self) -> Klein:
        return self._another


class DuplicateHasher:
    """
    Every L{DuplicateHasher} has the same hash value and compares equal to
    every other L{DuplicateHasher}.
    """

    __slots__ = ("_identifier",)

    def __init__(self, identifier: str) -> None:
        self._identifier = identifier

    myRouter = Klein()

    @myRouter.route("/")
    def root(self, request: IRequest) -> KleinRenderable:
        return self._identifier

    def __hash__(self) -> int:
        return 1

    def __eq__(self, other: Any) -> bool:
        return True


class KleinTestCase(unittest.TestCase):
    def test_route(self) -> None:
        """
        L{Klein.route} adds functions as routable endpoints.
        """
        app = Klein()

        @app.route("/foo")
        def foo(request: IRequest) -> KleinRenderable:
            return "foo"

        c = app.url_map.bind("foo")
        self.assertEqual(c.match("/foo"), ("foo", {}))
        self.assertEqual(len(app.endpoints), 1)

        self.assertEqual(app.execute_endpoint("foo", DummyRequest(1)), "foo")

    def test_mapByIdentity(self) -> None:
        """
        Routes are routed to the proper object regardless of its C{__hash__}
        implementation.
        """
        a = DuplicateHasher("a")
        b = DuplicateHasher("b")

        # Sanity check
        d = {}
        d[a] = "test"
        self.assertEqual(d.get(b), "test")

        self.assertEqual(
            a.myRouter.execute_endpoint("root", DummyRequest(1)), "a"
        )
        self.assertEqual(
            b.myRouter.execute_endpoint("root", DummyRequest(1)), "b"
        )

    def test_preserveIdentityWhenPossible(self) -> None:
        """
        Repeated accesses of the same L{Klein} attribute on the same instance
        should result in an identically bound instance, when possible.
        "Possible" is defined by a writable instance-level attribute named
        C{__klein_bound_<the name of the Klein attribute on the class>__}, and
        something is maintaining a strong reference to the L{Klein} instance.
        """

        # This is the desirable property.
        class DuplicateHasherWithWritableAttribute(DuplicateHasher):
            __slots__ = ("__klein_bound_myRouter__",)

        a = DuplicateHasherWithWritableAttribute("a")
        self.assertIs(a.myRouter, a.myRouter)

        b = DuplicateHasher("b")
        # The following is simply an unfortunate consequence of the
        # implementation choice here (i.e.: to insist on a specific writable
        # attribute), and could be changed (for example, by doing something
        # more elaborate with the identity of the object containing the
        # router).  However, checking this also sets a sort of bounded "worst
        # case" scenario"; it still works, nobody raises an exception, it's
        # just not identical.
        self.assertIsNot(b.myRouter, b.myRouter)

    def test_kleinNotFoundOnClass(self) -> None:
        """
        When the Klein object can't find itself on the class it still preserves
        identity.
        """

        class Wrap:
            def __init__(self, wrapped: Klein) -> None:
                self._wrapped = wrapped

            def __get__(self, instance: Any, owner: object) -> Any:
                if instance is None:
                    return self
                return self._wrapped.__get__(instance, owner)

        class TwoRouters:
            app1 = Wrap(Klein())
            app2 = Wrap(Klein())

        tr = TwoRouters()

        self.assertIs(tr.app1, tr.app1)
        self.assertIs(tr.app2, tr.app2)

    def test_bindInstanceIgnoresBlankProperties(self) -> None:
        """
        L{Klein.__get__} doesn't propagate L{AttributeError} when
        searching for the bound L{Klein} instance.
        """

        class ClassProperty:
            def __get__(self, instance: Any, owner: object) -> Any:
                raise AttributeError(
                    "you just don't have that special something"
                )

        class Oddment:
            __something__ = ClassProperty()
            app = Klein()

        self.assertIsInstance(Oddment().app, Klein)

    def test_submountedRoute(self) -> None:
        """
        L{Klein.subroute} adds functions as routable endpoints.
        """
        app = Klein()

        with app.subroute("/sub") as app:

            @app.route("/prefixed_uri")
            def foo_endpoint(request: IRequest) -> KleinRenderable:
                return b"foo"

        c = app.url_map.bind("sub/prefixed_uri")
        self.assertEqual(c.match("/sub/prefixed_uri"), ("foo_endpoint", {}))
        self.assertEqual(len(app.endpoints), 1)
        self.assertEqual(
            app.execute_endpoint("foo_endpoint", DummyRequest(1)), b"foo"
        )

    def test_stackedRoute(self) -> None:
        """
        L{Klein.route} can be stacked to create multiple endpoints of a single
        function.
        """
        app = Klein()

        @app.route("/foo")
        @app.route("/bar", endpoint="bar")
        def foobar(request: IRequest) -> KleinRenderable:
            return "foobar"

        self.assertEqual(len(app.endpoints), 2)

        c = app.url_map.bind("foo")
        self.assertEqual(c.match("/foo"), ("foobar", {}))
        self.assertEqual(
            app.execute_endpoint("foobar", DummyRequest(1)), "foobar"
        )

        self.assertEqual(c.match("/bar"), ("bar", {}))
        self.assertEqual(app.execute_endpoint("bar", DummyRequest(2)), "foobar")

    def test_branchRoute(self) -> None:
        """
        L{Klein.route} should create a branch path which consumes all children
        when the branch keyword argument is True.
        """
        app = Klein()

        @app.route("/foo/", branch=True)
        def branchfunc(request: IRequest) -> KleinRenderable:
            return "foo"

        c = app.url_map.bind("foo")
        self.assertEqual(c.match("/foo/"), ("branchfunc", {}))
        self.assertEqual(
            c.match("/foo/bar"), ("branchfunc_branch", {"__rest__": "bar"})
        )

        self.assertEqual(
            app.endpoints["branchfunc"].__name__,
            "route '/foo/' executor for branchfunc",
        )
        self.assertEqual(
            app.endpoints["branchfunc_branch"].__name__,
            "branch route '/foo/' executor for branchfunc",
        )
        self.assertEqual(
            app.execute_endpoint(
                "branchfunc_branch", DummyRequest("looking for foo")
            ),
            "foo",
        )

    def test_bindable(self) -> None:
        """
        L{bindable} is a decorator which allows a function decorated by @route
        to have a uniform signature regardless of whether it is receiving a
        bound object from its L{Klein} or not.
        """
        k = Klein()
        calls = []

        @k.route("/test")
        @bindable
        def method(self: Any, request: IRequest) -> KleinRenderable:
            calls.append((self, request))
            return "7"

        req = cast(IRequest, object())
        k.execute_endpoint("method", req)

        class BoundTo:
            app = k

        b = BoundTo()
        b.app.execute_endpoint("method", req)

        self.assertEqual(calls, [(None, req), (b, req)])
        self.assertEqual(originalName(method), "method")

    def test_modified(self) -> None:
        """
        L{modified} is a decorator which alters the thing that it decorates,
        and describes itself as such.
        """

        def annotate(decoratee: object) -> object:
            decoratee.supersized = True  # type: ignore[attr-defined]
            return decoratee

        def add(a: int, b: int) -> int:
            return a + b

        @modified("supersizer", add, modifier=annotate)
        def megaAdd(a: int, b: int) -> int:
            return add(a * 1000, b * 10000)

        self.assertEqual(megaAdd.supersized, True)
        self.assertEqual(add.supersized, True)  # type: ignore[attr-defined]
        self.assertIn("supersizer for add", str(megaAdd))
        self.assertEqual(megaAdd(3, 4), 43000)

    def test_classicalRoute(self) -> None:
        """
        L{Klein.route} may be used a method decorator when a L{Klein} instance
        is defined as a class variable.
        """
        bar_calls = []

        class Foo:
            app = Klein()

            @app.route("/bar")
            def bar(self, request: IRequest) -> KleinRenderable:
                bar_calls.append((self, request))
                return "bar"

        foo = Foo()
        c = foo.app.url_map.bind("bar")

        self.assertEqual(c.match("/bar"), ("bar", {}))
        self.assertEqual(
            foo.app.execute_endpoint("bar", DummyRequest(1)), "bar"
        )
        self.assertEqual(bar_calls, [(foo, DummyRequest(1))])

    def test_classicalRouteWithTwoInstances(self) -> None:
        """
        Multiple instances of a class with a L{Klein} attribute and
        L{Klein.route}'d methods can be created and their L{Klein}s used
        independently.
        """

        class Foo:
            app = Klein()

            def __init__(self) -> None:
                self.bar_calls: List[Tuple["Foo", IRequest]] = []

            @app.route("/bar")
            def bar(self, request: IRequest) -> KleinRenderable:
                self.bar_calls.append((self, request))
                return "bar"

        foo_1 = Foo()
        foo_1_app = foo_1.app
        foo_2 = Foo()
        foo_2_app = foo_2.app

        dr1 = DummyRequest(1)
        dr2 = DummyRequest(2)

        foo_1_app.execute_endpoint("bar", dr1)
        foo_2_app.execute_endpoint("bar", dr2)

        self.assertEqual(foo_1.bar_calls, [(foo_1, dr1)])
        self.assertEqual(foo_2.bar_calls, [(foo_2, dr2)])

    def test_classicalRouteWithBranch(self) -> None:
        """
        Multiple instances of a class with a L{Klein} attribute and
        L{Klein.route}'d methods can be created and their L{Klein}s used
        independently.
        """

        class Foo:
            app = Klein()

            def __init__(self) -> None:
                self.bar_calls: List[Tuple["Foo", IRequest]] = []

            @app.route("/bar/", branch=True)
            def bar(self, request: IRequest) -> KleinRenderable:
                self.bar_calls.append((self, request))
                return "bar"

        foo_1 = Foo()
        foo_1_app = foo_1.app
        foo_2 = Foo()
        foo_2_app = foo_2.app

        dr1 = DummyRequest(1)
        dr2 = DummyRequest(2)

        foo_1_app.execute_endpoint("bar_branch", dr1)
        foo_2_app.execute_endpoint("bar_branch", dr2)

        self.assertEqual(foo_1.bar_calls, [(foo_1, dr1)])
        self.assertEqual(foo_2.bar_calls, [(foo_2, dr2)])

    def test_branchDoesntRequireTrailingSlash(self) -> None:
        """
        L{Klein.route} should create a branch path which consumes all children,
        when the branch keyword argument is True and there is no trailing /
        on the path.
        """
        app = Klein()

        @app.route("/foo", branch=True)
        def foo(request: IRequest) -> KleinRenderable:
            return "foo"

        c = app.url_map.bind("foo")
        self.assertEqual(
            c.match("/foo/bar"), ("foo_branch", {"__rest__": "bar"})
        )

    @patch("klein._app.KleinResource")
    @patch("klein._app.Site")
    @patch("klein._app.log")
    @patch("klein._app.reactor")
    def test_run(
        self, reactor: Any, mock_log: Any, mock_site: Any, mock_kr: Any
    ) -> None:
        """
        L{Klein.run} configures a L{KleinResource} and a L{Site}
        listening on the specified interface and port, and logs
        to stdout.
        """
        app = Klein()

        app.run("localhost", 8080)

        reactor.listenTCP.assert_called_with(
            8080, mock_site.return_value, backlog=50, interface="localhost"
        )
        reactor.run.assert_called_with()

        mock_site.assert_called_with(mock_kr.return_value)
        mock_kr.assert_called_with(app)
        mock_log.startLogging.assert_called_with(stdout)

    @patch("klein._app.KleinResource")
    @patch("klein._app.Site")
    @patch("klein._app.log")
    @patch("klein._app.reactor")
    def test_runWithLogFile(
        self, reactor: Any, mock_log: Any, mock_site: Any, mock_kr: Any
    ) -> None:
        """
        L{Klein.run} logs to the specified C{logFile}.
        """
        app = Klein()

        logFile = Mock()
        app.run("localhost", 8080, logFile=logFile)

        reactor.listenTCP.assert_called_with(
            8080, mock_site.return_value, backlog=50, interface="localhost"
        )

        reactor.run.assert_called_with()

        mock_site.assert_called_with(mock_kr.return_value)
        mock_kr.assert_called_with(app)
        mock_log.startLogging.assert_called_with(logFile)

    @patch("klein._app.KleinResource")
    @patch("klein._app.log")
    @patch("klein._app.serverFromString")
    @patch("klein._app.reactor")
    def test_runTCP6(
        self, reactor: Any, mock_sfs: Any, mock_log: Any, mock_kr: Any
    ) -> None:
        """
        L{Klein.run} called with tcp6 endpoint description.
        """
        app = Klein()
        interface = "2001\\:0DB8\\:f00e\\:eb00\\:\\:1"
        spec = f"tcp6:8080:interface={interface}"
        app.run(endpoint_description=spec)
        reactor.run.assert_called_with()
        mock_sfs.assert_called_with(reactor, spec)
        mock_log.startLogging.assert_called_with(stdout)
        mock_kr.assert_called_with(app)

    @patch("klein._app.KleinResource")
    @patch("klein._app.log")
    @patch("klein._app.serverFromString")
    @patch("klein._app.reactor")
    def test_runSSL(
        self, reactor: Any, mock_sfs: Any, mock_log: Any, mock_kr: Any
    ) -> None:
        """
        L{Klein.run} called with SSL endpoint specification.
        """
        app = Klein()
        key = "key.pem"
        cert = "cert.pem"
        dh_params = "dhparam.pem"
        spec_template = "ssl:443:privateKey={0}:certKey={1}"
        spec = spec_template.format(key, cert, dh_params)
        app.run(endpoint_description=spec)
        reactor.run.assert_called_with()
        mock_sfs.assert_called_with(reactor, spec)
        mock_log.startLogging.assert_called_with(stdout)
        mock_kr.assert_called_with(app)

    @patch("klein._app.KleinResource")
    def test_resource(self, mock_kr: Any) -> None:
        """
        L{Klien.resource} returns a L{KleinResource}.
        """
        app = Klein()
        resource = app.resource()

        mock_kr.assert_called_with(app)
        self.assertEqual(mock_kr.return_value, resource)

    def test_urlFor(self) -> None:
        """L{Klein.urlFor} builds an URL for an endpoint with parameters"""

        app = Klein()

        @app.route("/user/<name>")
        def userpage(request: IRequest, name: str) -> KleinRenderable:
            return name

        @app.route("/post/<int:postid>", endpoint="bar")
        def foo(request: IRequest, postid: int) -> KleinRenderable:
            return str(postid)

        request = MockRequest(b"/user/john")
        self.assertEqual(
            app.execute_endpoint("userpage", request, "john"), "john"
        )
        self.assertEqual(
            app.execute_endpoint("bar", MockRequest(b"/post/123"), 123), "123"
        )

        request = MockRequest(b"/addr")
        self.assertEqual(
            app.urlFor(request, "userpage", {"name": "john"}), "/user/john"
        )

        request = MockRequest(b"/addr")
        self.assertEqual(
            app.urlFor(
                request, "userpage", {"name": "john"}, force_external=True
            ),
            "http://localhost:8080/user/john",
        )

        request = MockRequest(b"/addr", host=b"example.com", port=4321)
        self.assertEqual(
            app.urlFor(
                request, "userpage", {"name": "john"}, force_external=True
            ),
            "http://example.com:4321/user/john",
        )

        request = MockRequest(b"/addr")
        url = app.urlFor(
            request,
            "userpage",
            {"name": "john", "age": 29},
            append_unknown=True,
        )
        self.assertEqual(url, "/user/john?age=29")

        request = MockRequest(b"/addr")
        self.assertEqual(
            app.urlFor(request, "bar", {"postid": 123}), "/post/123"
        )

        request = MockRequest(b"/addr")
        request.requestHeaders.removeHeader(b"host")
        self.assertEqual(
            app.urlFor(request, "bar", {"postid": 123}), "/post/123"
        )

        request = MockRequest(b"/addr")
        request.requestHeaders.removeHeader(b"host")
        with self.assertRaises(ValueError):
            app.urlFor(request, "bar", {"postid": 123}, force_external=True)



# -*- test-case-name: klein.test.test_message -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Tests for L{klein._message}.
"""

from abc import ABC, abstractmethod
from typing import cast

from twisted.internet.defer import ensureDeferred

from .._imessage import IHTTPMessage
from .._message import FountAlreadyAccessedError, bytesToFount, fountToBytes
from ._trial import TestCase
from .not_hypothesis import binary, given


__all__ = ()


class FrozenHTTPMessageTestsMixIn(ABC):
    """
    Mix-In class for implementations of IHTTPMessage.
    """

    @classmethod
    @abstractmethod
    def messageFromBytes(cls, data: bytes = b"") -> IHTTPMessage:
        """
        Return a new instance of an L{IHTTPMessage} implementation using the
        given bytes as the message body.
        """

    @classmethod
    def messageFromFountFromBytes(cls, data: bytes = b"") -> IHTTPMessage:
        """
        Return a new instance of an L{IHTTPMessage} implementation using a
        fount containing the given bytes as the message body.
        """
        return cls.messageFromBytes(bytesToFount(data))

    def test_interface_message(self) -> None:
        """
        Message instance implements L{IHTTPMessage}.
        """
        message = self.messageFromBytes()
        cast(TestCase, self).assertProvides(IHTTPMessage, message)

    @given(binary())
    def test_bodyAsFountFromBytes(self, data: bytes) -> None:
        """
        C{bodyAsFount} returns a fount with the same bytes given to
        C{__init__}.
        """
        message = self.messageFromBytes(data)
        fount = message.bodyAsFount()
        body = cast(TestCase, self).successResultOf(
            ensureDeferred(fountToBytes(fount))
        )

        cast(TestCase, self).assertEqual(body, data)

    @given(binary())
    def test_bodyAsFountFromBytesTwice(self, data: bytes) -> None:
        """
        C{bodyAsFount} raises L{FountAlreadyAccessedError} if called more than
        once, when created from bytes.
        """
        message = self.messageFromBytes(data)
        message.bodyAsFount()
        cast(TestCase, self).assertRaises(
            FountAlreadyAccessedError, message.bodyAsFount
        )

    @given(binary())
    def test_bodyAsFountFromFount(self, data: bytes) -> None:
        """
        C{bodyAsBytes} returns the bytes from the fount given to C{__init__}.
        """
        message = self.messageFromFountFromBytes(data)
        fount = message.bodyAsFount()
        body = cast(TestCase, self).successResultOf(
            ensureDeferred(fountToBytes(fount))
        )
        cast(TestCase, self).assertEqual(body, data)

    @given(binary())
    def test_bodyAsFountFromFountTwice(self, data: bytes) -> None:
        """
        C{bodyAsFount} raises L{FountAlreadyAccessedError} if called more than
        once, when created from a fount.
        """
        message = self.messageFromFountFromBytes(data)
        message.bodyAsFount()
        cast(TestCase, self).assertRaises(
            FountAlreadyAccessedError, message.bodyAsFount
        )

    @given(binary())
    def test_bodyAsBytesFromBytes(self, data: bytes) -> None:
        """
        C{bodyAsBytes} returns the same bytes given to C{__init__}.
        """
        message = self.messageFromBytes(data)
        body = cast(TestCase, self).successResultOf(
            ensureDeferred(message.bodyAsBytes())
        )
        cast(TestCase, self).assertEqual(body, data)

    @given(binary())
    def test_bodyAsBytesFromBytesCached(self, data: bytes) -> None:
        """
        C{bodyAsBytes} called twice returns the same object both times, when
        created from bytes.
        """
        message = self.messageFromBytes(data)
        body1 = cast(TestCase, self).successResultOf(
            ensureDeferred(message.bodyAsBytes())
        )
        body2 = cast(TestCase, self).successResultOf(
            ensureDeferred(message.bodyAsBytes())
        )
        cast(TestCase, self).assertIdentical(body1, body2)

    @given(binary())
    def test_bodyAsBytesFromFount(self, data: bytes) -> None:
        """
        C{bodyAsBytes} returns the bytes from the fount given to C{__init__}.
        """
        message = self.messageFromFountFromBytes(data)
        body = cast(TestCase, self).successResultOf(
            ensureDeferred(message.bodyAsBytes())
        )
        cast(TestCase, self).assertEqual(body, data)

    @given(binary())
    def test_bodyAsBytesFromFountCached(self, data: bytes) -> None:
        """
        C{bodyAsBytes} called twice returns the same object both times, when
        created from a fount.
        """
        message = self.messageFromFountFromBytes(data)
        body1 = cast(TestCase, self).successResultOf(
            ensureDeferred(message.bodyAsBytes())
        )
        body2 = cast(TestCase, self).successResultOf(
            ensureDeferred(message.bodyAsBytes())
        )
        cast(TestCase, self).assertIdentical(body1, body2)



# -*- test-case-name: klein.test.test_request_compat -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Tests for L{klein._irequest}.
"""

import functools
from string import ascii_uppercase
from types import MappingProxyType
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Mapping,
    Sequence,
    TypeVar,
)

from hyperlink import DecodedURL, EncodedURL

from twisted.internet.defer import ensureDeferred
from twisted.web.iweb import IRequest

from .._headers import IHTTPHeaders
from .._message import FountAlreadyAccessedError
from .._request import IHTTPRequest
from .._request_compat import HTTPRequestWrappingIRequest
from ._trial import TestCase
from .not_hypothesis import binary, decoded_urls, given, text
from .test_resource import MockRequest


__all__ = ()


emptyMapping: Mapping[Any, Any] = MappingProxyType({})

_T = TypeVar("_T")
_R = TypeVar("_R")


def ensuringDeferred(
    fn: Callable[[_T], Coroutine[Any, Any, _R]]
) -> Callable[[_T], Awaitable[_R]]:
    @functools.wraps(fn)
    def wrapper(self: _T) -> Awaitable[_R]:
        return ensureDeferred(fn(self))

    return wrapper


class HTTPRequestWrappingIRequestTests(TestCase):
    """
    Tests for L{HTTPRequestWrappingIRequest}.
    """

    def legacyRequest(
        self,
        path: bytes = b"/",
        method: bytes = b"GET",
        host: bytes = b"localhost",
        port: int = 8080,
        isSecure: bool = False,
        body: bytes = b"",
        headers: Mapping[bytes, Sequence[bytes]] = emptyMapping,
    ) -> IRequest:
        return MockRequest(
            path=path,
            method=method,
            host=host,
            port=port,
            isSecure=isSecure,
            body=body,
            headers=headers,
        )

    def test_interface(self) -> None:
        """
        L{HTTPRequestWrappingIRequest} implements L{IHTTPRequest}.
        """
        request = HTTPRequestWrappingIRequest(request=self.legacyRequest())
        self.assertProvides(IHTTPRequest, request)

    @given(text(alphabet=ascii_uppercase, min_size=1))
    def test_method(self, methodText: str) -> None:
        """
        L{HTTPRequestWrappingIRequest.method} matches the underlying legacy
        request method.
        """
        legacyRequest = self.legacyRequest(method=methodText.encode("ascii"))
        request = HTTPRequestWrappingIRequest(request=legacyRequest)
        self.assertEqual(request.method, methodText)

    @given(decoded_urls())
    def test_uri(self, url: DecodedURL) -> None:
        """
        L{HTTPRequestWrappingIRequest.uri} matches the underlying legacy
        request URI.
        """
        uri = url.asURI()  # Normalize as (computer-friendly) URI

        assert uri.port is not None  # Tells mypy it's not an Optional

        path = (
            uri.replace(scheme="", host="", port=None).asText().encode("ascii")
        )
        legacyRequest = self.legacyRequest(
            isSecure=(uri.scheme == "https"),
            host=uri.host.encode("ascii"),
            port=uri.port,
            path=path,
        )
        request = HTTPRequestWrappingIRequest(request=legacyRequest)

        uriNormalized = uri
        requestURINormalized = request.uri.asURI()

        # Needed because non-equal URLs can render as the same strings
        def strURL(url: EncodedURL) -> str:
            return (
                "URL(scheme={url.scheme!r}, "
                "userinfo={url.userinfo!r}, "
                "host={url.host!r}, "
                "port={url.port!r}, "
                "path={url.path!r}, "
                "query={url.query!r}, "
                "fragment={url.fragment!r}, "
                "rooted={url.rooted})"
            ).format(url=url)

        self.assertEqual(
            requestURINormalized,
            uriNormalized,
            "{} != {}".format(
                strURL(requestURINormalized), strURL(uriNormalized)
            ),
        )

    def test_headers(self) -> None:
        """
        L{HTTPRequestWrappingIRequest.headers} returns an
        L{HTTPRequestWrappingIRequest} containing the underlying legacy request
        headers.
        """
        legacyRequest = self.legacyRequest()
        request = HTTPRequestWrappingIRequest(request=legacyRequest)
        self.assertProvides(IHTTPHeaders, request.headers)

    def test_bodyAsFountTwice(self) -> None:
        """
        L{HTTPRequestWrappingIRequest.bodyAsFount} raises
        L{FountAlreadyAccessedError} if called more than once.
        """
        legacyRequest = self.legacyRequest()
        request = HTTPRequestWrappingIRequest(request=legacyRequest)
        request.bodyAsFount()
        self.assertRaises(FountAlreadyAccessedError, request.bodyAsFount)

    @given(binary())
    def test_bodyAsBytes(self, data: bytes) -> None:
        """
        L{HTTPRequestWrappingIRequest.bodyAsBytes} matches the underlying
        legacy request body.
        """
        legacyRequest = self.legacyRequest(body=data)
        request = HTTPRequestWrappingIRequest(request=legacyRequest)
        body = self.successResultOf(ensureDeferred(request.bodyAsBytes()))

        self.assertEqual(body, data)

    @ensuringDeferred
    async def test_bodyAsBytesCached(self) -> None:
        """
        L{HTTPRequestWrappingIRequest.bodyAsBytes} called twice returns the
        same object both times.
        """
        data = b"some data"
        legacyRequest = self.legacyRequest(body=data)
        request = HTTPRequestWrappingIRequest(request=legacyRequest)
        body1 = await request.bodyAsBytes()
        body2 = await request.bodyAsBytes()

        self.assertIdentical(body1, body2)



from typing import Any

from zope.interface import Interface
from zope.interface.verify import verifyObject

from twisted.trial.unittest import SynchronousTestCase

from klein.interfaces import ISession, ISessionStore, SessionMechanism
from klein.storage.memory import MemorySessionStore, declareMemoryAuthorizer


class IFoo(Interface):
    """
    Testing interface 1.
    """


class IBar(Interface):
    """
    Testing interface 2.
    """


class MemoryTests(SynchronousTestCase):
    """
    Tests for memory-based session storage.
    """

    def test_interfaceCompliance(self) -> None:
        """
        Verify that the session store complies with the relevant interfaces.
        """
        store = MemorySessionStore()
        verifyObject(ISessionStore, store)
        verifyObject(
            ISession,
            self.successResultOf(
                store.newSession(True, SessionMechanism.Header)
            ),
        )

    def test_noAuthorizers(self) -> None:
        """
        By default, L{MemorySessionStore} contains no authorizers and the
        sessions it returns will authorize any supplied interfaces as None.
        """
        store = MemorySessionStore()
        session = self.successResultOf(
            store.newSession(True, SessionMechanism.Header)
        )
        self.assertEqual(
            self.successResultOf(session.authorize([IFoo, IBar])), {}
        )

    def test_simpleAuthorization(self) -> None:
        """
        L{MemorySessionStore.fromAuthorizers} takes a set of functions
        decorated with L{declareMemoryAuthorizer} and constructs a session
        store that can authorize for those interfaces.
        """

        @declareMemoryAuthorizer(IFoo)
        def fooMe(interface: Any, session: Any, componentized: Any) -> int:
            return 1

        @declareMemoryAuthorizer(IBar)
        def barMe(interface: Any, session: Any, componentized: Any) -> int:
            return 2

        store = MemorySessionStore.fromAuthorizers([fooMe, barMe])
        session = self.successResultOf(
            store.newSession(False, SessionMechanism.Cookie)
        )
        self.assertEqual(
            self.successResultOf(session.authorize([IBar, IFoo])),
            {IFoo: 1, IBar: 2},
        )



# -*- test-case-name: klein.test.test_form -*-

import json
from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    NoReturn,
    Optional,
    Sequence,
    Type,
    cast,
)

import attr
from zope.interface import Attribute, Interface, implementer

from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.python.components import Componentized, registerAdapter
from twisted.web.error import MissingRenderMethod
from twisted.web.http import FORBIDDEN
from twisted.web.iweb import IRenderable, IRequest
from twisted.web.resource import Resource
from twisted.web.template import Element, Tag, TagLoader, tags

from ._app import KleinRenderable, _call
from ._decorators import bindable
from .interfaces import (
    EarlyExit,
    IDependencyInjector,
    IRequestLifecycle,
    IRequiredParameter,
    ISession,
    SessionMechanism,
    ValidationError,
    ValueAbsent,
)


class CrossSiteRequestForgery(Resource):
    """
    Cross site request forgery detected.  Request aborted.
    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = message

    def render(self, request: IRequest) -> bytes:
        """
        For all HTTP methods, return a 403.
        """
        request.setResponseCode(FORBIDDEN, b"FAILURECSRF")
        return ("CSRF TOKEN FAILURE: " + self.message).encode("utf-8")


CSRF_PROTECTION = "__csrf_protection__"


def textConverter(value: AnyStr) -> str:
    """
    Converter for form values (which may be any type of string) into text.
    """
    if isinstance(value, bytes):
        return value.decode("utf-8")
    else:
        return value


class IParsedJSONBody(Interface):
    """
    Marker interface for the dict parsed from the request body's JSON contents.
    """

    # TODO: how to allow applications to pass options to loads, such as
    # parse_float?


@implementer(IRequiredParameter)
@attr.s(auto_attribs=True, frozen=True)
class Field:
    """
    A L{Field} is a static part of a L{Form}.

    @ivar converter: The converter.
    """

    converter: Callable[[AnyStr], Any]
    formInputType: str
    pythonArgumentName: Optional[str] = None
    formFieldName: Optional[str] = None
    formLabel: Optional[str] = None
    default: Optional[Any] = attr.ib(default=None, cmp=False)
    required: bool = True
    noLabel: bool = False
    value: str = ""
    error: Optional[ValidationError] = None

    # IRequiredParameter
    def registerInjector(
        self,
        injectionComponents: Componentized,
        parameterName: str,
        requestLifecycle: IRequestLifecycle,
    ) -> IDependencyInjector:
        """
        Register this form field as a dependency injector.
        """
        protoForm = IProtoForm(injectionComponents)
        return cast(
            IDependencyInjector,
            protoForm.addField(self.maybeNamed(parameterName)),
        )

    def maybeNamed(self, name: str) -> "Field":
        """
        Create a new L{Field} like this one, but with all the name default
        values filled in.

        @param name: the name.
        """

        def maybe(
            it: Optional[str], that: Optional[str] = name
        ) -> Optional[str]:
            return that if it is None else it

        return attr.evolve(
            self,
            pythonArgumentName=maybe(self.pythonArgumentName),
            formFieldName=maybe(self.formFieldName),
            formLabel=maybe(
                self.formLabel, name.capitalize() if not self.noLabel else None
            ),
        )

    def asTags(self) -> Iterable[Tag]:
        """
        Convert this L{Field} into some stuff that can be rendered in a
        L{twisted.web.template}.

        @return: A new set of tags to include in a template.
        """
        value = self.value
        if value is None:
            value = ""  # type: ignore[unreachable]
        fieldName = self.formFieldName

        if fieldName is None:
            raise ValueError("Cannot generate tags for unnamed form field.")

        input_tag = tags.input(
            type=self.formInputType,
            name=fieldName,
            value=value,
        )
        error_tags = []
        if self.error:
            error_tags.append(
                tags.div(class_="klein-form-validation-error")(
                    self.error.message  # type: ignore[arg-type]
                )
            )
        if self.formLabel:
            yield tags.label(self.formLabel, ": ", input_tag, *error_tags)
        else:
            yield input_tag
            yield from error_tags

    def extractValue(self, request: IRequest) -> Any:
        """
        Extract a value from the request.

        In the case of key/value form posts, this attempts to reliably make the
        value into str.  In the case of a JSON post, however, it will simply
        extract the value from the top-level dictionary, which means it could
        be any arrangement of JSON-serializiable objects.
        """
        fieldName = self.formFieldName
        if fieldName is None:
            raise ValueError("Cannot extract unnamed form field.")
        contentType = request.getHeader(b"content-type")
        if contentType is not None and contentType.startswith(
            b"application/json"
        ):
            # TODO: parse only once, please.
            parsed = cast(Componentized, request).getComponent(IParsedJSONBody)
            if parsed is None:
                request.content.seek(0)
                octets = request.content.read()
                characters = octets.decode("utf-8")
                parsed = json.loads(characters)
                cast(Componentized, request).setComponent(
                    IParsedJSONBody, parsed
                )
            if fieldName not in parsed:
                return None
            return parsed[fieldName]
        allValues = request.args.get(fieldName.encode("utf-8"))
        if allValues:
            return allValues[0].decode("utf-8")
        else:
            return None

    def validateValue(self, value: Any) -> Any:
        """
        Validate the given text and return a converted Python object to use, or
        fail with L{ValidationError}.

        @param value: The value that was extracted by L{Field.extractValue}.

        @return: The converted value.
        """
        if value is None:
            if self.required:
                raise ValueAbsent("a value was required but none was supplied")
            else:
                return self.default
        try:
            return self.converter(value)
        except ValueError as ve:
            raise ValidationError(str(ve))

    @classmethod
    def text(cls, **kw: Any) -> "Field":
        """
        Shorthand for a form field that contains a short string, and will be
        rendered as a plain <input>.
        """
        return cls(converter=textConverter, formInputType="text", **kw)

    @classmethod
    def password(cls, **kw: Any) -> "Field":
        """
        Shorthand for a form field that, like L{text}, contains a short string,
        but should be obscured when typed (and, to the extent possible,
        obscured in other sensitive contexts, such as logging.)
        """
        return cls(converter=textConverter, formInputType="password", **kw)

    @classmethod
    def hidden(cls, name: str, value: str, **kw: Any) -> "Field":
        """
        Shorthand for a hidden field.
        """
        return cls(
            converter=textConverter,
            formInputType="hidden",
            noLabel=True,
            value=value,
            **kw,
        ).maybeNamed(name)

    @classmethod
    def number(
        cls,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        kind: Type = float,
        **kw: Any,
    ) -> "Field":
        """
        An integer within the range [minimum, maximum].
        """

        def bounded_number(text: AnyStr) -> Any:
            try:
                value = kind(text)
            except (ValueError, ArithmeticError):
                raise ValidationError("not a valid number")
            else:
                if minimum is not None and value < minimum:
                    raise ValidationError("value must be >=" + repr(minimum))
                if maximum is not None and value > maximum:
                    raise ValidationError("value must be <=" + repr(maximum))
                return value

        return cls(converter=bounded_number, formInputType="number", **kw)

    @classmethod
    def submit(cls, value: str) -> "Field":
        """
        A field representing a submit button, with a value (displayed on the
        button).
        """
        return cls(
            converter=textConverter,
            formInputType="submit",
            noLabel=True,
            default=value,
        )


@implementer(IRenderable)
@attr.s(auto_attribs=True)
class RenderableForm:
    """
    An L{IRenderable} representing a renderable form.

    @ivar prevalidationValues: a L{dict} mapping {L{Field}: L{list} of
        L{str}}, representing the value that each field received as part of
        the request.

    @ivar validationErrors: a L{dict} mapping {L{Field}: L{ValidationError}}
    """

    _form: "IForm"
    _session: ISession
    _action: str
    _method: str
    _enctype: str
    _encoding: str
    prevalidationValues: Dict[Field, str] = attr.ib(factory=dict)
    validationErrors: Dict[Field, ValidationError] = attr.ib(factory=dict)

    ENCTYPE_FORM_DATA = "multipart/form-data"
    ENCTYPE_URL_ENCODED = "application/x-www-form-urlencoded"

    def _fieldForCSRF(self) -> Field:
        """
        @return: A hidden L{Field} containing the cross-site request forgery
            protection token.
        """
        return Field.hidden(CSRF_PROTECTION, self._session.identifier)

    def _fieldsToRender(self) -> Iterable[Field]:
        """
        @return: an iterable of L{Field} objects to include in the HTML
            representation of this form.  This includes:

                - all the user-specified fields in the form

                - the CSRF protection hidden field

                - if no "submit" buttons are included in the form, one
                  additional field for a default submit button so the form can
                  be submitted.
        """
        anySubmit = False
        for field in self._form.fields:
            yield attr.evolve(
                field,
                value=self.prevalidationValues.get(field, field.value),
                error=self.validationErrors.get(field, None),
            )
            if field.formInputType == "submit":
                anySubmit = True
        if not anySubmit:
            yield Field(
                converter=str,
                formInputType="submit",
                value="submit",
                formFieldName="__klein_auto_submit__",
            )
        if self._method.lower() == "post":
            yield self._fieldForCSRF()

    # Public interface below.

    def lookupRenderMethod(self, name: str) -> NoReturn:
        """
        Form renderers don't supply any render methods, so this just always
        raises L{MissingRenderMethod}.
        """
        raise MissingRenderMethod(self, name)

    def render(self, request: IRequest) -> Tag:  # type: ignore[override]
        """
        Render this form to the given request.
        """
        formAttributes = {
            "accept-charset": self._encoding,
            "class": "klein-form",
        }
        if self._method.lower() == "post":
            # Enctype has no meaning on method="GET" forms.
            formAttributes.update(enctype=self._enctype)
        return tags.form(
            action=self._action, method=self._method, **formAttributes
        )(field.asTags() for field in self._fieldsToRender())

    def glue(self) -> List[Tag]:
        """
        Provide any glue necessary to render this form; this must be dropped
        into the template within the C{<form>} tag.

        Presently, this glue includes only the CSRF token argument, but Klein
        reserves the right to add arbitrary HTML here.  This should not create
        any user-visible content, however.

        @return: some HTML elements in the form of renderable objects for
            L{twisted.web.template}
        """
        return list(self._fieldForCSRF().asTags())


@bindable
def defaultValidationFailureHandler(
    instance: Optional[object],
    request: IRequest,
    fieldValues: "FieldValues",
) -> Element:
    """
    This is the default validation failure handler, which will be used by form
    handlers (i.e. any routes which use L{klein.Requirer} to require a field)
    in the case of any input validation failure when no other validation
    failure handler is registered via L{Form.onValidationFailureFor}.

    Its behavior is to simply return an HTML rendering of the form object,
    which includes inline information about fields which failed to validate.

    @param instance: The instance associated with the router that the form
        handler was handled on.
    @param request: The request including the form submission.
    @return: Any object acceptable from a Klein route.
    """
    session = cast(Componentized, request).getComponent(ISession)
    request.setResponseCode(400)
    enctype = (
        (
            request.getHeader(b"content-type")
            or RenderableForm.ENCTYPE_URL_ENCODED.encode("ascii")
        )
        .split(b";")[0]
        .decode("charmap")
    )
    renderable = RenderableForm(
        fieldValues.form,
        session,
        "/".join(
            segment.decode("utf-8", errors="replace")
            for segment in request.prepath
        ),
        request.method,
        enctype,
        "utf-8",
        fieldValues.prevalidationValues,
        fieldValues.validationErrors,
    )

    return Element(TagLoader(renderable))


_requirerFunctionWithForm = Any
_routeCallable = Any


class IProtoForm(Interface):
    """
    Marker interface for L{ProtoForm}.
    """

    fields: Sequence[Field] = Attribute("Form fields")

    def addField(field: Field) -> "FieldInjector":
        """
        Add the given field to the form ultimately created here.
        """


class IForm(Interface):
    """
    Marker interface for form attached to dependency injection components.
    """

    fields: Sequence[Field] = Attribute("Form fields")

    def populateRequestValues(
        injectionComponents: Componentized,
        instance: Any,
        request: IRequest,
    ) -> Deferred:
        """
        Extract the values present in this request and populate a
        L{FieldValues} object.
        """


@implementer(IProtoForm)
@attr.s(auto_attribs=True)
class ProtoForm:
    """
    Form-builder.
    """

    _componentized: Componentized
    _lifecycle: IRequestLifecycle
    fields: List[Field] = attr.ib(factory=list)

    @classmethod
    def fromComponentized(cls, componentized: Componentized) -> "ProtoForm":
        """
        Create a ProtoForm from a componentized object.
        """
        rl = IRequestLifecycle(componentized)
        assert rl is not None
        return cls(componentized, rl)

    def addField(self, field: Field) -> "FieldInjector":
        self.fields.append(field)
        return FieldInjector(self._componentized, field, self._lifecycle)


class IFieldValues(Interface):
    """
    Marker interface for parsed fields.
    """

    form: IForm = Attribute("Form")
    arguments: Dict[str, Any] = Attribute("Arguments")
    prevalidationValues: Dict[Field, Optional[str]] = Attribute(
        "Pre-validation values"
    )
    validationErrors: Dict[Field, ValidationError] = Attribute(
        "Validation errors"
    )

    def validate(instance: Any, request: IRequest) -> Deferred:
        """
        If any validation errors have occurred, raise a relevant exception.
        """


@implementer(IFieldValues)
@attr.s(auto_attribs=True)
class FieldValues:
    """
    Reified post-parsing values for HTTP form submission.
    """

    form: "Form"
    arguments: Dict[str, Any]
    prevalidationValues: Dict[Field, str]
    validationErrors: Dict[Field, ValidationError]
    _injectionComponents: Componentized

    @inlineCallbacks
    def validate(
        self, instance: Any, request: IRequest
    ) -> Generator[Any, object, None]:
        if self.validationErrors:
            result = cast(
                KleinRenderable,
                (
                    yield _call(
                        instance,
                        IValidationFailureHandler(
                            self._injectionComponents,
                            defaultValidationFailureHandler,
                        ),
                        request,
                        self,
                    )
                ),
            )
            raise EarlyExit(result)


@implementer(IDependencyInjector)
@attr.s(auto_attribs=True)
class FieldInjector:
    """
    Field injector.
    """

    _componentized: Componentized
    _field: Field
    _lifecycle: IRequestLifecycle

    def injectValue(
        self, instance: Any, request: IRequest, routeParams: Dict[str, Any]
    ) -> Any:
        """
        Inject the given value into the form.
        """
        assert self._field.pythonArgumentName is not None
        return IFieldValues(request).arguments.get(
            self._field.pythonArgumentName
        )

    def finalize(self) -> None:
        """
        Finalize this ProtoForm into a real form.
        """
        if IForm(self._componentized, None) is not None:
            return

        finalForm = cast(IForm, Form(IProtoForm(self._componentized).fields))
        self._componentized.setComponent(IForm, finalForm)

        # XXX set requiresComponents argument here to ISession if CSRF is
        # required; add flag somewhere to indicate if a form is
        # side-effect-free (like a search field) that can be handled even
        # without a CSRF token.
        @bindable
        def populateValuesHook(instance: Any, request: IRequest) -> Deferred:
            return finalForm.populateRequestValues(
                self._componentized, instance, request
            )

        self._lifecycle.addPrepareHook(
            populateValuesHook,
            provides=[IFieldValues],
            requires=[ISession],
        )


registerAdapter(ProtoForm.fromComponentized, Componentized, IProtoForm)


class IValidationFailureHandler(Interface):
    """
    Validation failure handler callable interface.
    """

    def __call__(request: IRequest) -> KleinRenderable:
        """
        Called to handle a validation failure.
        """


def checkCSRF(request: IRequest) -> None:
    """
    Check the request for cross-site request forgery, raising an EarlyExit if
    it is found.
    """
    # TODO: optionalize CSRF protection for GET forms
    session = ISession(request, None)
    token = None
    if request.method in (b"GET", b"HEAD"):
        # Idempotent requests don't require CRSF validation.  (Don't have
        # destructive GETs or bad stuff will happen to you in general!)
        return
    if session is not None:
        if session.authenticatedBy == SessionMechanism.Header:
            # CSRF is possible because browsers automatically send cookies when
            # foreign domains provoke them into sending requests.  If a request
            # is authenticated by a header, then no checking is necessary.
            return
        # We have a session, we weren't authenticated by a header... time to
        # check that token.
        token = request.args.get(CSRF_PROTECTION.encode("ascii"), [b""])[
            0
        ].decode("ascii")
        if token == session.identifier:
            # The token matches.  We're OK.
            return
    # leak only the value passed, not the actual token, just in
    # case there's some additional threat vector there
    raise EarlyExit(CrossSiteRequestForgery(f"Invalid CSRF token: {token!r}"))


@implementer(IForm)
@attr.s(auto_attribs=True, hash=False)
class Form:
    """
    A L{Form} is a collection of fields attached to a function.
    """

    fields: Sequence[Field]

    @staticmethod
    def onValidationFailureFor(
        handler: _requirerFunctionWithForm,
    ) -> Callable[[Callable], Callable]:
        """
        Register a function to be run in the event of a validation failure for
        the input to a particular form handler.

        Generally used like so::

            requirer = Requirer(...)
            @requirer.prerequisite([ISession])
            def procureSession(request):
                return SessionProcurer(...).procureSession(request)
            router = Klein()
            @requirer.require(router.route("/", methods=['POST']),
                              someField=Field.text())
            def formHandler(someField):
                ...
            # Handle input validation failures for handleForm
            @Form.onValidationFailureFor(formHandler)
            def handleValidationFailures(request, fieldValues):
                return "Your inputs didn't validate."

        @see: L{defaultValidationFailureHandler} for a more detailed
            description of the decorated function's expected signature.  The
            additional parameter it is passed is a L{FieldValues} instance.

        @param handler: The form handler - i.e. function decorated by
            L{Form.handler} - for which the decorated function will handle
            validation failures.

        @return: a decorator that decorates a function with the signature
            C{(request, form) -> thing klein can render}.
        """

        def decorate(decoratee: Callable) -> Callable:
            handler.injectionComponents.setComponent(
                IValidationFailureHandler, decoratee
            )
            return decoratee

        return decorate

    @inlineCallbacks
    def populateRequestValues(
        self,
        injectionComponents: Componentized,
        instance: Any,
        request: IRequest,
    ) -> Generator[Any, object, None]:
        assert IFieldValues(request, None) is None

        validationErrors = {}
        prevalidationValues = {}
        arguments = {}

        checkCSRF(request)

        for field in self.fields:
            text = field.extractValue(request)
            prevalidationValues[field] = text
            try:
                value = field.validateValue(text)
                argName = field.pythonArgumentName
                if argName is None:
                    raise ValidationError("Form fields must all have names.")
            except ValidationError as ve:
                validationErrors[field] = ve
            else:
                arguments[argName] = value
        values = FieldValues(
            self,
            arguments,
            prevalidationValues,
            validationErrors,
            injectionComponents,
        )
        yield values.validate(instance, request)
        cast(Componentized, request).setComponent(IFieldValues, values)

    @classmethod
    def rendererFor(
        cls,
        decoratedFunction: _requirerFunctionWithForm,
        action: str,
        method: str = "POST",
        enctype: str = RenderableForm.ENCTYPE_FORM_DATA,
        encoding: str = "utf-8",
    ) -> "RenderableFormParam":
        """
        A form parameter that can render a form declared as a number of fields
        on another route.

        Use like so::

            class MyFormApp:
                router = Klein()
                requirer = Requirer()

                @requirer.require(
                    router.route("/handle-form", methods=["POST"]),
                    name=Field.text(), value=Field.integer(),
                )
                def formRoute(self, name, value):
                    ...

                @requirer.require(
                    router.route("/show-form", methods=["GET"]),
                    form=Form.rendererFor(formRoute)
                )
                def showForm(self, form):
                    return form

        As a L{RenderableForm} provides L{IRenderable}, you may return the
        parameter directly
        """
        form: Optional[IForm] = IForm(
            decoratedFunction.injectionComponents, None
        )
        if form is None:
            form = Form([])
        return RenderableFormParam(form, action, method, enctype, encoding)


@implementer(IRequiredParameter, IDependencyInjector)
@attr.s(auto_attribs=True)
class RenderableFormParam:
    """
    A L{RenderableFormParam} implements L{IRequiredParameter} and
    L{IDependencyInjector} to provide a L{RenderableForm} to your route.
    """

    _form: IForm
    _action: str
    _method: str
    _enctype: str
    _encoding: str

    def registerInjector(
        self,
        injectionComponents: Componentized,
        parameterName: str,
        requestLifecycle: IRequestLifecycle,
    ) -> "RenderableFormParam":
        return self

    def injectValue(
        self, instance: Any, request: IRequest, routeParams: Dict[str, Any]
    ) -> RenderableForm:
        """
        Create the renderable form from the request.
        """
        return RenderableForm(
            self._form,
            ISession(request),
            self._action,
            self._method,
            self._enctype,
            self._encoding,
            prevalidationValues={},
            validationErrors={},
        )

    def finalize(self) -> None:
        """
        Nothing to do upon finalization.
        """



"""
We have had a history of U{bad
experiences<https://github.com/twisted/klein/issues/561>} with Hypothesis in
Klein, and maybe it's not actually a good application of this tool at all.  As
such we have removed it, at least for now.  This module presents a vaguely
Hypothesis-like stub, to keep the structure of our tests in a
Hypothesis-friendly shape, in case we want to put it back.
"""

from functools import wraps
from itertools import product
from string import ascii_uppercase
from typing import Callable, Iterable, Optional, Tuple, TypeVar

from hyperlink import DecodedURL


T = TypeVar("T")
S = TypeVar("S")


def given(
    *args: Callable[[], Iterable[T]],
    **kwargs: Callable[[], Iterable[T]],
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    def decorator(testMethod: Callable[..., None]) -> Callable[..., None]:
        @wraps(testMethod)
        def realTestMethod(self: S) -> None:
            everyPossibleArgs = product(
                *[eachFactory() for eachFactory in args]
            )
            everyPossibleKwargs = product(
                *[
                    [(name, eachValue) for eachValue in eachFactory()]
                    for (name, eachFactory) in kwargs.items()
                ]
            )
            everyPossibleSignature = product(
                everyPossibleArgs, everyPossibleKwargs
            )
            # not quite the _full_ cartesian product but the whole point is
            # that we're making a feeble attempt at this rather than bringing
            # in hypothesis.
            for computedArgs, computedPairs in everyPossibleSignature:
                computedKwargs = dict(computedPairs)
                testMethod(self, *computedArgs, **computedKwargs)

        return realTestMethod

    return decorator


def binary() -> Callable[[], Iterable[bytes]]:
    """
    Generate some binary data.
    """

    def params() -> Iterable[bytes]:
        return [b"data", b"data data data", b"\x00" * 50, b""]

    return params


def ascii_text(min_size: int) -> Callable[[], Iterable[str]]:
    """
    Generate some ASCII strs.
    """

    def params() -> Iterable[str]:
        yield from [
            "ascii-text",
            "some more ascii text",
        ]
        assert min_size, "nothing needs 0-length strings right now"

    return params


def latin1_text(min_size: int = 0) -> Callable[[], Iterable[str]]:
    """
    Generate some strings encodable as latin1
    """

    def params() -> Iterable[str]:
        yield from [
            "latin1-text",
            "some more latin1 text",
            "hére is latin1 text",
        ]
        if not min_size:
            yield ""

    return params


def text(
    min_size: int = 0, alphabet: Optional[str] = None
) -> Callable[[], Iterable[str]]:
    """
    Generate some text.
    """

    def params() -> Iterable[str]:
        if alphabet == ascii_uppercase:
            yield from ascii_text(min_size)()
            return
        yield from latin1_text(min_size)()
        yield "\N{SNOWMAN}"

    return params


def textHeaderPairs() -> Callable[[], Iterable[Iterable[Tuple[str, str]]]]:
    """
    Generate some pairs of headers with text values.
    """

    def params() -> Iterable[Iterable[Tuple[str, str]]]:
        return [[], [("text", "header")]]

    return params


def bytesHeaderPairs() -> Callable[[], Iterable[Iterable[Tuple[str, bytes]]]]:
    """
    Generate some pairs of headers with bytes values.
    """

    def params() -> Iterable[Iterable[Tuple[str, bytes]]]:
        return [[], [("bytes", b"header")]]

    return params


def booleans() -> Callable[[], Iterable[bool]]:
    def parameters() -> Iterable[bool]:
        yield True
        yield False

    return parameters


def jsonObjects() -> Callable[[], Iterable[object]]:
    def parameters() -> Iterable[object]:
        yield {}
        yield {"hello": "world"}
        yield {"here is": {"some": "nesting"}}
        yield {
            "and": "multiple",
            "keys": {
                "with": "nesting",
                "and": 1234,
                "numbers": ["with", "lists", "too"],
                "also": ("tuples", "can", "serialize"),
            },
        }

    return parameters


def decoded_urls() -> Callable[[], Iterable[DecodedURL]]:
    """
    Generate a few URLs U{with only path and domain names
    <https://github.com/python-hyper/hyperlink/issues/181>} kind of like
    Hyperlink's own hypothesis strategy.
    """

    def parameters() -> Iterable[DecodedURL]:
        yield DecodedURL.from_text("https://example.com/")
        yield DecodedURL.from_text("https://example.com")
        yield DecodedURL.from_text("http://example.com/")
        yield DecodedURL.from_text("https://example.com/é")
        yield DecodedURL.from_text("https://súbdomain.example.com/ascii/path/")

    return parameters



from ._iform import ValidationError, ValueAbsent
from ._interfaces import IKleinRequest
from ._isession import (
    EarlyExit,
    IDependencyInjector,
    IRequestLifecycle,
    IRequiredParameter,
    ISession,
    ISessionProcurer,
    ISessionStore,
    ISimpleAccount,
    ISimpleAccountBinding,
    NoSuchSession,
    SessionMechanism,
    TooLateForCookies,
    TransactionEnded,
)


__all__ = (
    "EarlyExit",
    "IDependencyInjector",
    "IKleinRequest",
    "IRequestLifecycle",
    "IRequiredParameter",
    "ISession",
    "ISessionProcurer",
    "ISessionStore",
    "ISimpleAccount",
    "ISimpleAccountBinding",
    "NoSuchSession",
    "SessionMechanism",
    "TooLateForCookies",
    "TransactionEnded",
    "ValidationError",
    "ValueAbsent",
)



"""
This file contains code that should validate with type checking if type hints
are correct.

The code is not executed, it is here just to be checked by mypy.
"""

from twisted.internet.defer import succeed
from twisted.web.iweb import IRequest
from twisted.web.resource import Resource
from twisted.web.template import Element, Tag

from klein import Klein, KleinRenderable


class Application:
    router = Klein()

    @router.route("/object")
    def returnsObject(self, request: IRequest) -> KleinRenderable:
        return object()  # type: ignore[return-value]

    @router.route("/str")
    def returnsStr(self, request: IRequest) -> KleinRenderable:
        return ""

    @router.route("/bytes")
    def returnsBytes(self, request: IRequest) -> KleinRenderable:
        return b""

    @router.route("/iresource")
    def returnsIResource(self, request: IRequest) -> KleinRenderable:
        return Resource()

    @router.route("/irenderable")
    def returnsIRenderable(self, request: IRequest) -> KleinRenderable:
        return Element()

    @router.route("/tag")
    def returnsTag(self, request: IRequest) -> KleinRenderable:
        return Tag("")

    @router.route("/none")
    def returnsNone(self, request: IRequest) -> KleinRenderable:
        return None

    @router.route("/deferred-object")
    def returnsDeferredObject(self, request: IRequest) -> KleinRenderable:
        return succeed(object())  # type: ignore[arg-type]

    @router.route("/deferred-str")
    def returnsDeferredStr(self, request: IRequest) -> KleinRenderable:
        return succeed("")

    @router.route("/deferred-bytes")
    def returnsDeferredBytes(self, request: IRequest) -> KleinRenderable:
        return succeed(b"")

    @router.route("/deferred-iresource")
    def returnsDeferredIResource(self, request: IRequest) -> KleinRenderable:
        return succeed(Resource())

    @router.route("/deferred-irenderable")
    def returnsDeferredIRenderable(self, request: IRequest) -> KleinRenderable:
        return succeed(Element())

    @router.route("/deferred-tag")
    def returnsDeferredTag(self, request: IRequest) -> KleinRenderable:
        return succeed(Tag(""))

    @router.route("/deferred-none")
    def returnsDeferredNone(self, request: IRequest) -> KleinRenderable:
        return succeed(None)



from typing import TYPE_CHECKING

from ._app import (
    Klein,
    KleinErrorHandler,
    KleinRenderable,
    KleinRouteHandler,
    handle_errors,
    route,
    run,
    subroute,
    url_for,
    urlFor,
)
from ._dihttp import RequestComponent, RequestURL, Response
from ._form import Field, FieldValues, Form, RenderableForm
from ._plating import Plating
from ._requirer import Requirer
from ._session import Authorization, SessionProcurer
from ._version import __version__ as _incremental_version


if TYPE_CHECKING:
    # Inform mypy of import shenanigans.
    from .resource import _SpecialModuleObject

    resource = _SpecialModuleObject(None)
else:
    from . import resource

__all__ = (
    "Klein",
    "KleinErrorHandler",
    "KleinRenderable",
    "KleinRouteHandler",
    "Plating",
    "Field",
    "FieldValues",
    "Form",
    "RequestComponent",
    "RequestURL",
    "Response",
    "RenderableForm",
    "SessionProcurer",
    "Authorization",
    "Requirer",
    "__author__",
    "__copyright__",
    "__license__",
    "__version__",
    "handle_errors",
    "resource",
    "route",
    "run",
    "subroute",
    "urlFor",
    "url_for",
)


# Make it a str, for backwards compatibility
__version__ = _incremental_version.base()

__author__ = "The Klein contributors (see AUTHORS)"
__license__ = "MIT"
__copyright__ = f"Copyright 2011-2021 {__author__}"



# Copyright (c) 2011-2021. See LICENSE for details.

"""
Extensions to L{twisted.trial}.
"""

from typing import Type

from zope.interface import Interface
from zope.interface.exceptions import Invalid
from zope.interface.verify import verifyObject

from twisted.trial.unittest import SynchronousTestCase


__all__ = ()


class TestCase(SynchronousTestCase):
    """
    Extensions to L{SynchronousTestCase}.
    """

    def assertProvides(self, interface: Type[Interface], obj: object) -> None:
        """
        Assert that a object provides an interface.

        @param interface: The interface the object is expected to provide.
        @param obj: The object to test.
        """
        try:
            self.assertTrue(verifyObject(interface, obj))
        except Invalid:
            self.fail(f"{obj} does not provide {interface}")



# -*- test-case-name: klein.test.test_request -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Tests for L{klein._request}.
"""

from hyperlink import DecodedURL

from .._headers import FrozenHTTPHeaders
from .._imessage import IHTTPMessage
from .._request import FrozenHTTPRequest, IHTTPRequest
from ._trial import TestCase
from .test_message import FrozenHTTPMessageTestsMixIn


__all__ = ()


class FrozenHTTPRequestTests(FrozenHTTPMessageTestsMixIn, TestCase):
    """
    Tests for L{FrozenHTTPRequest}.
    """

    @staticmethod
    def requestFromBytes(data: bytes = b"") -> FrozenHTTPRequest:
        return FrozenHTTPRequest(
            method="GET",
            uri=DecodedURL.fromText("https://twistedmatrix.com/"),
            headers=FrozenHTTPHeaders(rawHeaders=()),
            body=data,
        )

    @classmethod
    def messageFromBytes(cls, data: bytes = b"") -> IHTTPMessage:
        return cls.requestFromBytes(data)

    def test_interface(self) -> None:
        """
        L{FrozenHTTPRequest} implements L{IHTTPRequest}.
        """
        request = self.requestFromBytes()
        self.assertProvides(IHTTPRequest, request)

    def test_initInvalidBodyType(self) -> None:
        """
        L{FrozenHTTPRequest} raises L{TypeError} when given a body of an
        unknown type.
        """
        e = self.assertRaises(
            TypeError,
            FrozenHTTPRequest,
            method="GET",
            uri=DecodedURL.fromText("https://twistedmatrix.com/"),
            headers=FrozenHTTPHeaders(rawHeaders=()),
            body=object(),
        )
        self.assertEqual(str(e), "body must be bytes or IFount")



# Copyright (c) 2011-2021. See LICENSE for details.

"""
Internal interface definitions.

All Zope Interface classes should be imported from here so that type checking
works, since mypy doesn't otherwise get along with Zope Interface.
"""

from typing import Mapping, Optional, Union

from zope.interface import Attribute, Interface


KleinQueryValue = Union[str, int, float]


class IKleinRequest(Interface):
    branch_segments = Attribute("Segments consumed by a branch route.")
    mapper = Attribute("L{werkzeug.routing.MapAdapter}")

    def url_for(
        endpoint: str,
        values: Optional[Mapping[str, KleinQueryValue]] = None,
        method: Optional[str] = None,
        force_external: bool = False,
        append_unknown: bool = True,
    ) -> str:
        """
        L{werkzeug.routing.MapAdapter.build}
        """


__all__ = ()



class ValidationError(Exception):
    """
    A L{ValidationError} is raised by L{Field.extractValue}.
    """

    def __init__(self, message: object) -> None:
        """
        Initialize a L{ValidationError} with a message to show to the user.
        """
        super().__init__(message)
        self.message = message


class ValueAbsent(ValidationError):
    """
    A value was required but none was supplied.
    """



"""
Tests for exported API.
"""

from twisted.trial import unittest


class PublicSymbolsTestCase(unittest.TestCase):
    """
    Tests for public API modules.
    """

    def test_klein(self) -> None:
        """
        Test exports from L{klein}.
        """
        import klein as k
        import klein._app as a
        import klein._plating as p

        self.assertIdentical(k.Klein, a.Klein)
        self.assertIdentical(k.handle_errors, a.handle_errors)
        self.assertIdentical(k.route, a.route)
        self.assertIdentical(k.run, a.run)

        self.assertIdentical(k.Plating, p.Plating)

    def test_klein_resource(self) -> None:
        """
        Test export of C{resource} from L{klein}.
        """
        import klein as k
        import klein._app as a

        self.assertIdentical(k.resource()._app, a.resource()._app)

    def test_app(self) -> None:
        """
        Test exports from L{klein.app}.
        """
        import klein._app as _a
        import klein.app as a

        self.assertIdentical(a.Klein, _a.Klein)
        self.assertIdentical(a.KleinRequest, _a.KleinRequest)
        self.assertIdentical(a.handle_errors, _a.handle_errors)
        self.assertIdentical(a.route, _a.route)
        self.assertIdentical(a.run, _a.run)

    def test_interfaces(self) -> None:
        """
        Test exports from L{klein.interfaces}.
        """
        import klein._interfaces as _i
        import klein.interfaces as i

        self.assertIdentical(i.IKleinRequest, _i.IKleinRequest)

    def test_resource(self) -> None:
        """
        Test exports from L{klein.resource}.
        """
        import klein._resource as _r
        import klein.resource as r

        self.assertIdentical(r.KleinResource, _r.KleinResource)
        self.assertIdentical(r.ensure_utf8_bytes, _r.ensure_utf8_bytes)



"""
Tests for L{klein.plating}.
"""


import json
from typing import Any

import attr

from twisted.internet.defer import Deferred, succeed
from twisted.trial.unittest import SynchronousTestCase
from twisted.trial.unittest import TestCase as AsynchronousTestCase
from twisted.web.error import FlattenerError, MissingRenderMethod
from twisted.web.template import slot, tags

from .. import Klein, Plating
from .._plating import ATOM_TYPES, PlatedElement, resolveDeferredObjects
from .not_hypothesis import booleans, given, jsonObjects
from .test_resource import MockRequest, _render


page = Plating(
    defaults={
        "title": "default title unchanged",
        Plating.CONTENT: "NEVER MIND THE CONTENT",
    },
    tags=tags.html(
        tags.head(tags.title(slot("title"))),
        tags.body(
            tags.h1(slot("title")),
            tags.div(slot(Plating.CONTENT), Class="content"),
            tags.div(id="rendermethod", render="registeredRenderMethod"),
        ),
    ),
)

element = Plating(
    defaults={
        "a": "NO VALUE FOR A",
        "b": "NO VALUE FOR B",
    },
    tags=tags.div(
        tags.span("a: ", slot("a")),
        tags.span("b: ", slot("b")),
    ),
)


@element.widgeted
def enwidget(a, b):
    """
    Provide some values for the L{element} template.
    """
    return {"a": a, "b": b}


@element.widgeted
def deferredEnwidget(a, b):
    """
    Provide some L{Deferred} values for the L{element} template.
    """
    return {"a": succeed(a), "b": succeed(b)}


@page.renderMethod
def registeredRenderMethod(*args):
    """
    Register a render method with the page so that the template can use it.
    """
    # Variable signature because since it's for testing, 'page' here has the
    # unusual property that is used in both bound (i.e. "has self") and unbound
    # contexts.
    tag = args[-1]
    return tag("(self)" if len(args) == 3 else "", "some text!")


class InstanceWidget:
    """
    A class with a method that's a L{Plating.widget}.
    """

    @element.widgeted
    def enwidget(self, a, b):
        """
        Provide some values for the L{element} template.
        """
        return {"a": a, "b": b}

    @element.widgeted
    def deferredEnwidget(self, a, b):
        """
        Provide some L{Deferred} values for the L{element} template.
        """
        return {"a": succeed(a), "b": succeed(b)}


@attr.s(auto_attribs=True)
class DeferredValue:
    """
    A value within a JSON serializable object that is deferred.

    @param value: The value.

    @param deferred: The L{Deferred} representing the value.
    """

    value: Any
    deferred: Deferred = attr.ib(factory=Deferred)

    def resolve(self):
        """
        Resolve the L{Deferred} that represents the value with the
        value itself.
        """
        self.deferred.callback(self.value)


def transformJSONObject(jsonObject, transformer):
    """
    Recursively apply a transforming function to a JSON serializable
    object, returning a new, transformed object.

    @param json_object: A JSON serializable object to transform.

    @param transformer: A one-argument callable that will be applied
        to each member of the object.

    @return: A transformed copy of C{jsonObject}.
    """

    def visit(obj):
        """
        Recur through the object, sometimes replacing values with
        L{Deferred}s
        """
        if isinstance(obj, ATOM_TYPES):
            return transformer(obj)
        elif isinstance(obj, tuple):
            return tuple(transformer(child) for child in obj)
        elif isinstance(obj, list):
            return [transformer(child) for child in obj]
        elif isinstance(obj, dict):
            return {transformer(k): transformer(v) for k, v in obj.items()}
        else:
            raise AssertionError(f"Object of unknown type {obj!r}")

    return visit(jsonObject)


class TransformJSONObjectTests(SynchronousTestCase):
    """
    Tests for L{transform_json_object}.
    """

    def test_transform_atom(self):
        """
        L{transform_json_object} transforms a representative atomic
        JSON data types.
        """

        def inc(value):
            return value + 1

        self.assertEqual(transformJSONObject(1, inc), 2)

    def test_transform_tuple(self):
        """
        L{transformJSONObject} transforms L{tuple}s.
        """

        def inc(value):
            return value + 1

        self.assertEqual(transformJSONObject((1, 1), inc), (2, 2))

    def test_transform_list(self):
        """
        L{transformJSONObject} transforms L{list}s.
        """

        def inc(value):
            return value + 1

        self.assertEqual(transformJSONObject([1, 1], inc), [2, 2])

    def test_transform_dict(self):
        """
        L{transformJSONObject} transforms L{dict}s.
        """

        def inc(value):
            return value + 1

        self.assertEqual(transformJSONObject({2: 2}, inc), {3: 3})

    def test_transform_unserializable(self):
        """
        L{transformJSONObject} will not transform objects that are
        not JSON serializable.
        """
        self.assertRaises(
            AssertionError, transformJSONObject, set(), list.append
        )


class ResolveDeferredObjectsTests(SynchronousTestCase):
    """
    Tests for L{resolve_deferred_objects}.
    """

    @given(
        jsonObject=jsonObjects(),
        shouldWrapDeferred=booleans(),
    )
    def test_resolveObjects(
        self, jsonObject: object, shouldWrapDeferred: bool
    ) -> None:
        """
        A JSON serializable object that may contain L{Deferred}s or a
        L{Deferred} that resolves to a JSON serializable object
        resolves to an object that contains no L{Deferred}s.
        """
        deferredValues = []

        def maybeWrapInDeferred(value):
            if shouldWrapDeferred:
                deferredValues.append(DeferredValue(value))
                return deferredValues[-1].deferred
            else:
                return value

        deferredJSONObject = transformJSONObject(
            jsonObject,
            maybeWrapInDeferred,
        )

        resolved: "Deferred[Any]" = resolveDeferredObjects(deferredJSONObject)

        for value in deferredValues:
            value.resolve()

        self.assertEqual(self.successResultOf(resolved), jsonObject)

    @given(
        jsonObject=jsonObjects(),
        shouldInjectElements=booleans(),
    )
    def test_elementSerialized(
        self, jsonObject: object, shouldInjectElements: bool
    ) -> None:
        """
        A L{PlatedElement} within a JSON serializable object replaced
        by its JSON representation.
        """

        def injectPlatingElements(value):
            if shouldInjectElements and isinstance(value, dict):
                return PlatedElement(
                    slot_data=value,
                    preloaded=tags.html(),
                    boundInstance=None,
                    presentationSlots={},
                    renderers={},
                )
            else:
                return value

        withPlatingElements = transformJSONObject(
            jsonObject,
            injectPlatingElements,
        )

        resolved: "Deferred[Any]" = resolveDeferredObjects(withPlatingElements)

        self.assertEqual(self.successResultOf(resolved), jsonObject)

    def test_unserializableObject(self):
        """
        An object that cannot be serialized causes the L{Deferred} to
        fail with an informative L{TypeError}.

        """

        @attr.s
        class ConsistentRepr:
            """
            Objects with a predictable repr
            """

        exception = self.failureResultOf(
            resolveDeferredObjects(ConsistentRepr())
        ).value
        self.assertIsInstance(exception, TypeError)
        self.assertIn("ConsistentRepr() not JSON serializable", str(exception))


class PlatingTests(AsynchronousTestCase):
    """
    Tests for L{Plating}.
    """

    def setUp(self):
        """
        Create an app and a resource wrapping that app for this test.
        """
        self.app = Klein()
        self.kr = self.app.resource()

    def get(self, uri):
        """
        Issue a virtual GET request to the given path that is expected to
        succeed synchronously, and return the generated request object and
        written bytes.
        """
        request = MockRequest(uri)
        d = _render(self.kr, request)
        self.successResultOf(d)
        return request, request.getWrittenData()

    def test_template_html(self):
        """
        Rendering a L{Plating.routed} decorated route results in templated
        HTML.
        """

        @page.routed(self.app.route("/"), tags.span(slot("ok")))
        def plateMe(request):
            return {"ok": "test-data-present"}

        request, written = self.get(b"/")

        self.assertIn(b"<span>test-data-present</span>", written)
        self.assertIn(b"<title>default title unchanged</title>", written)

    def test_selfhood(self):
        """
        Rendering a L{Plating.routed} decorated route on a method still results
        in the decorated method receiving the appropriate C{self}.
        """

        class AppObj:
            app = Klein()

            def __init__(self, x):
                self.x = x

            @page.routed(app.route("/"), tags.span(slot("yeah")))
            def plateInstance(self, request):
                return {"yeah": "test-instance-data-" + self.x}

        obj = AppObj("confirmed")
        self.kr = obj.app.resource()
        request, written = self.get(b"/")

        self.assertIn(b"<span>test-instance-data-confirmed</span>", written)
        self.assertIn(b"<title>default title unchanged</title>", written)
        self.assertIn(b'<div id="rendermethod">(self)some text!</div>', written)

    def test_template_json(self):
        """
        Rendering a L{Plating.routed} decorated route with a query parameter
        asking for JSON will yield JSON instead.
        """

        @page.routed(self.app.route("/"), tags.span(slot("ok")))
        def plateMe(request):
            return {"ok": "an-plating-test"}

        request, written = self.get(b"/?json=true")
        self.assertEqual(
            request.responseHeaders.getRawHeaders(b"content-type")[0],
            b"application/json",
        )
        self.assertEqual(
            {"ok": "an-plating-test", "title": "default title unchanged"},
            json.loads(written.decode("utf-8")),
        )

    def test_template_json_contains_deferred(self):
        """
        Rendering a L{Plating.routed} decorated route with a query
        parameter asking for JSON waits until the L{Deferred}s
        returned by the route have fired.
        """

        @page.routed(self.app.route("/"), tags.span(slot("ok")))
        def plateMe(request):
            return {"ok": succeed("an-plating-test")}

        request, written = self.get(b"/?json=true")
        self.assertEqual(
            request.responseHeaders.getRawHeaders(b"content-type")[0],
            b"application/json",
        )
        self.assertEqual(
            {"ok": "an-plating-test", "title": "default title unchanged"},
            json.loads(written.decode("utf-8")),
        )

    def test_template_numbers(self):
        """
        Data returned from a plated method may include numeric types (integers,
        floats, and possibly longs), which although they are not normally
        serializable by twisted.web.template, will be converted by plating into
        their decimal representation.
        """

        @page.routed(
            self.app.route("/"),
            tags.div(
                tags.span(slot("anInteger")),
                tags.i(slot("anFloat")),
                tags.b(slot("anLong")),
            ),
        )
        def plateMe(result):
            return {
                "anInteger": 7,
                "anFloat": 3.2,
                "anLong": 0x10000000000000001,
            }

        request, written = self.get(b"/")

        self.assertIn(b"<span>7</span>", written)
        self.assertIn(b"<i>3.2</i>", written)
        self.assertIn(b"<b>18446744073709551617</b>", written)

    def test_render_list(self):
        """
        The C{:list} renderer suffix will render the slot named by the renderer
        as a list, filling each slot.
        """

        @page.routed(
            self.app.route("/"),
            tags.ul(tags.li(slot("item"), render="subplating:list")),
        )
        def rsrc(request):
            return {"subplating": [1, 2, 3]}

        request, written = self.get(b"/")

        self.assertIn(b"<ul><li>1</li><li>2</li><li>3</li></ul>", written)
        self.assertIn(b"<title>default title unchanged</title>", written)

    def test_widget_function(self):
        """
        A function decorated with L{Plating.wigeted} can be directly
        invoked.
        """
        self.assertEqual(enwidget(5, 6), {"a": 5, "b": 6})
        self.assertEqual(InstanceWidget().enwidget(7, 8), {"a": 7, "b": 8})

    def test_renderMethod(self):
        """
        L{Plating.renderMethod} registers a renderer that may be used by the
        L{Plating}'s template.
        """

        @page.routed(self.app.route("/"), [])
        def rsrc(request):
            return {}

        request, written = self.get(b"/")
        self.assertIn(b'<div id="rendermethod">some text!</div>', written)

    def test_widget_html(self):
        """
        When L{Plating.widgeted} is applied as a decorator, it gives the
        decorated function a C{widget} attribute which is a version of the
        function with a modified return type that turns it into a renderable
        HTML sub-element that may fill a slot.
        """

        @page.routed(
            self.app.route("/"),
            tags.div(
                tags.div(slot("widget")), tags.div(slot("instance-widget"))
            ),
        )
        def rsrc(request):
            return {
                "widget": enwidget.widget(a=3, b=4),
                "instance-widget": InstanceWidget().enwidget.widget(5, 6),
            }

        request, written = self.get(b"/")

        self.assertIn(b"<span>a: 3</span>", written)
        self.assertIn(b"<span>b: 4</span>", written)
        self.assertIn(b"<span>a: 5</span>", written)
        self.assertIn(b"<span>b: 6</span>", written)

    def test_widget_json(self):
        """
        When L{Plating.widgeted} is applied as a decorator, and the result is
        serialized to JSON, it appears the same as the returned value despite
        the HTML-friendly wrapping described above.
        """

        @page.routed(
            self.app.route("/"),
            tags.div(
                tags.div(slot("widget")), tags.div(slot("instance-widget"))
            ),
        )
        def rsrc(request):
            return {
                "widget": enwidget.widget(a=3, b=4),
                "instance-widget": InstanceWidget().enwidget.widget(5, 6),
            }

        request, written = self.get(b"/?json=1")
        self.assertEqual(
            json.loads(written.decode("utf-8")),
            {
                "widget": {"a": 3, "b": 4},
                "instance-widget": {"a": 5, "b": 6},
                "title": "default title unchanged",
            },
        )

    def test_widget_json_deferred(self):
        """
        When L{Plating.widgeted} is applied as a decorator, and the result is
        serialized to JSON, it appears the same as the returned value despite
        the HTML-friendly wrapping described above.
        """

        @page.routed(
            self.app.route("/"),
            tags.div(
                tags.div(slot("widget")), tags.div(slot("instance-widget"))
            ),
        )
        def rsrc(request):
            instance = InstanceWidget()
            return {
                "widget": deferredEnwidget.widget(a=3, b=4),
                "instance-widget": instance.deferredEnwidget.widget(5, 6),
            }

        request, written = self.get(b"/?json=1")
        self.assertEqual(
            json.loads(written.decode("utf-8")),
            {
                "widget": {"a": 3, "b": 4},
                "instance-widget": {"a": 5, "b": 6},
                "title": "default title unchanged",
            },
        )

    def test_prime_directive_return(self):
        """
        Nothing within these Articles Of Federation shall authorize the United
        Federation of Planets to alter the return value of a callable by
        applying a decorator to it...
        """
        exact_result = {"ok": "some nonsense value"}

        @page.routed(self.app.route("/"), tags.span(slot("ok")))
        def plateMe(request):
            return exact_result

        self.assertIdentical(plateMe(None), exact_result)

    def test_prime_directive_arguments(self):
        """
        ... or shall require the function to modify its signature under these
        Articles Of Federation.
        """

        @page.routed(self.app.route("/"), tags.span(slot("ok")))
        def plateMe(request, one, two, three):
            return (one, two, three)

        exact_one = {"one": "and"}
        exact_two = {"two": "and"}
        exact_three = {"three": "and"}
        result_one, result_two, result_three = plateMe(
            None, exact_one, exact_two, three=exact_three
        )

        self.assertIdentical(result_one, exact_one)
        self.assertIdentical(result_two, exact_two)
        self.assertIdentical(result_three, exact_three)

    def test_presentation_only_json(self):
        """
        Slots marked as "presentation only" will not be reflected in the
        output.
        """
        plating = Plating(
            tags=tags.span(slot("title")), presentation_slots={"title"}
        )

        @plating.routed(self.app.route("/"), tags.span(slot("data")))
        def justJson(request):
            return {"title": "uninteresting", "data": "interesting"}

        request, written = self.get(b"/?json=1")

        self.assertEqual(
            json.loads(written.decode("utf-8")), {"data": "interesting"}
        )

    def test_missing_renderer(self):
        """
        Missing renderers will result in an exception during rendering.
        """

        def test(missing):
            plating = Plating(tags=tags.span(slot(Plating.CONTENT)))

            @plating.routed(
                self.app.route("/"), tags.span(tags.span(render=missing))
            )
            def no(request):
                return {}

            self.get(b"/")
            [fe] = self.flushLoggedErrors(FlattenerError)
            self.assertIsInstance(fe.value.args[0], MissingRenderMethod)

        test("garbage")
        test("garbage:missing")



from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Sequence, Type

import attr
from constantly import NamedConstant, Names
from zope.interface import Attribute, Interface

from twisted.internet.defer import Deferred
from twisted.python.components import Componentized
from twisted.web.iweb import IRequest


if TYPE_CHECKING:
    from ._app import KleinRenderable


class NoSuchSession(Exception):
    """
    No such session could be found.
    """


class TooLateForCookies(Exception):
    """
    It's too late to set a cookie.
    """


class TransactionEnded(Exception):
    """
    Exception raised when.
    """


class SessionMechanism(Names):
    """
    Mechanisms which can be used to identify and authenticate a session.

    @cvar Cookie: The Cookie session mechanism involves looking up the session
        identifier via an HTTP cookie.  Session objects retrieved via this
        mechanism may be vulnerable to U{CSRF attacks
        <https://www.owasp.org/index.php/Cross-Site_Request_Forgery_(CSRF)>}
        and therefore must have CSRF protections applied to them.

    @cvar Header: The Header mechanism retrieves the session identifier via a
        separate header such as C{"X-Auth-Token"}.  Since a different-origin
        site in a browser can easily send a form submission including cookies,
        but I{can't} easily put stuff into other arbitrary headers, this does
        not require additional protections.
    """

    Cookie = NamedConstant()
    Header = NamedConstant()


class ISession(Interface):
    """
    An L{ISession} provider contains an identifier for the session, information
    about how the session was negotiated with the client software, and
    """

    identifier = Attribute(
        """
        L{str} identifying a session.

        This value should be:

            1. I{unique} - no two sessions have the same identifier

            2. I{unpredictable} - no one but the receipient of the session
               should be able to guess what it is

            3. I{opaque} - it should contain no interesting information
        """
    )

    isConfidential = Attribute(
        """
        A L{bool} indicating whether this session mechanism transmitted over an
        encrypted transport, i.e., HTTPS.  If C{True}, this means that this
        session can be used for sensitive information; otherwise, the
        information contained in it should be considered to be available to
        attackers.
        """
    )

    authenticatedBy = Attribute(
        """
        A L{SessionMechanism} indicating what mechanism was used to
        authenticate this session.
        """
    )

    def authorize(interfaces: Iterable[Type[Interface]]) -> Deferred:
        """
        Retrieve other objects from this session.

        This method is how you can retrieve application-specific objects from
        the general-purpose session; define interfaces for each facet of
        something accessible to a session, then pass it here and to the
        L{ISessionStore} implementation you're using.

        @param interfaces: A list of interfaces.

        @return: all of the providers that could be retrieved from the session.
        @rtype: L{Deferred} firing with L{dict} mapping
            L{zope.interface.interfaces.IInterface} to providers of each
            interface.  Interfaces which cannot be authorized will not be
            present as keys in this dictionary.
        """


class ISessionStore(Interface):
    """
    Backing storage for sessions.
    """

    def newSession(
        isConfidential: bool,
        authenticatedBy: SessionMechanism,
    ) -> Deferred:
        """
        Create a new L{ISession}.

        @return: a new session with a new identifier.
        @rtype: L{Deferred} firing with L{ISession}.
        """

    def loadSession(
        identifier: str,
        isConfidential: bool,
        authenticatedBy: SessionMechanism,
    ) -> Deferred:
        """
        Load a session given the given identifier and security properties.

        As an optimization for session stores where the back-end can generate
        session identifiers when the presented one is not found in the same
        round-trip to a data store, this method may return a L{Session} object
        with an C{identifier} attribute that does not match C{identifier}.
        However, please keep in mind when implementing L{ISessionStore} that
        this behavior is only necessary for requests where C{authenticatedBy}
        is L{SessionMechanism.Cookie}; an unauthenticated
        L{SessionMechanism.Header} session is from an API client and its
        session should be valid already.

        @return: an existing session with the given identifier.
        @rtype: L{Deferred} firing with L{ISession} or failing with
            L{NoSuchSession}.
        """

    def sentInsecurely(identifiers: Sequence[str]) -> None:
        """
        The transport layer has detected that the given identifiers have been
        sent over an unauthenticated transport.
        """


class ISimpleAccountBinding(Interface):
    """
    Data-store agnostic account / session binding manipulation API for "simple"
    accounts - i.e. those using username, password, and email address as a
    method to authenticate a user.

    This goes into a user-authentication-capable L{ISession} object's C{data}
    attribute as a component.
    """

    def bindIfCredentialsMatch(username: str, password: str) -> None:
        """
        Attach the session this is a component of to an account with the given
        username and password, if the given username and password correctly
        authenticate a principal.
        """

    def boundAccounts() -> Deferred:
        """
        Retrieve the accounts currently associated with the session this is a
        component of.

        @return: L{Deferred} firing with a L{list} of L{ISimpleAccount}.
        """

    def unbindThisSession() -> None:
        """
        Disassociate the session this is a component of from any accounts it's
        logged in to.
        """

    def createAccount(username: str, email: str, password: str) -> None:
        """
        Create a new account with the given username, email and password.
        """


class ISimpleAccount(Interface):
    """
    Data-store agnostic account interface.
    """

    username = Attribute(
        """
        Unicode username.
        """
    )

    accountID = Attribute(
        """
        Unicode account-ID.
        """
    )

    def bindSession(session: ISession) -> None:
        """
        Bind the given session to this account; i.e. authorize the given
        session to act on behalf of this account.
        """

    def changePassword(newPassword: str) -> None:
        """
        Change the password of this account.
        """


class ISessionProcurer(Interface):
    """
    An L{ISessionProcurer} wraps an L{ISessionStore} and can procure sessions
    that store, given HTTP request objects.
    """

    def procureSession(
        request: IRequest, forceInsecure: bool = False
    ) -> Deferred:
        """
        Retrieve a session using whatever technique is necessary.

        If the request already identifies an existing session in the store,
        retrieve it.  If not, create a new session and retrieve that.

        @param request: The request to procure a session from.
        @param forceInsecure: Even if the request was transmitted securely
            (i.e. over HTTPS), retrieve the session that would be used by the
            same browser if it were sending an insecure (i.e. over HTTP)
            request; by default, this is False, and the session's security will
            match that of the request.

        @return: a L{Deferred} that:

                - fires with an L{ISession} provider if the request describes
                  an existing, valid session, or, if the intersection of the
                  data in the request and the configuration of this
                  L{ISessionProcurer} allow for a cookie to be set immediately,
                  or

                - fails with L{NoSuchSession} if the request is unable to
                  negotiate a session based on the current request: this is
                  generally if the client is trying to use header-based
                  authentication (and therefore does not want a new cookie set)
                  or if this procurer is configured not to automatically create
                  new sessions on the fly, or

                - fails with L{TooLateForCookies} if the request bound to this
                  procurer has already sent the headers and therefore we can no
                  longer set a cookie, and we need to set a cookie.
        """


class IDependencyInjector(Interface):
    """
    An injector for a given dependency.
    """

    def injectValue(
        instance: Any, request: IRequest, routeParams: Dict[str, Any]
    ) -> Any:
        """
        Return a value to be injected into the parameter name specified by the
        IRequiredParameter.  This may return a Deferred, or an object, or an
        object directly providing the relevant interface.

        @param instance: The instance to which the Klein router processing this
            request is bound.

        @param request: The request being processed.

        @param routeParams: A (read-only) copy of the the arguments passed to
            the route by the layer below dependency injection (for example, URL
            parameters).
        """

    def finalize() -> None:
        """
        Finalize this injector before allowing the route to be created.

        Finalization generally includes:

            - attaching any hooks to the request lifecycle object that need to
              be run before/after each request

            - attaching any finalized component objects to the
              injectionComponents originally passed along to the
              IRequiredParameter that created this IDependencyInjector.
        """


class IRequestLifecycle(Interface):
    """
    Interface for adding hooks to the phases of a request's lifecycle.
    """

    def addPrepareHook(
        beforeHook: Callable,
        requires: Sequence[Type[Interface]] = (),
        provides: Sequence[Type[Interface]] = (),
    ) -> None:
        """
        Add a hook that promises to prepare the request by supplying the given
        interfaces as components on the request, and requires the given
        requirements.

        Prepare hooks are run I{before any} L{IDependencyInjector}s I{inject
        their values}.
        """


class IRequiredParameter(Interface):
    """
    A declaration that a given Python parameter is required to satisfy a given
    dependency at request-handling time.
    """

    def registerInjector(
        injectionComponents: Componentized,
        parameterName: str,
        lifecycle: IRequestLifecycle,
    ) -> IDependencyInjector:
        """
        Register the given injector at method-decoration time, informing it of
        its Python parameter name.

        @note: this happens at I{route definition} time, after all other
            injectors have been registered by
            L{IRequiredParameter.registerInjector}.

        @param lifecycle: An L{IRequestLifecycle} provider which contains hooks
            that will be run before and after each request.  If this injector
            has shared per-request dependencies that need to be executed before
            or after the request is processed, this method should attach them
            to those lists.  These hooks are supplied here rather than relying
            on C{injectValue} to run the requisite logic each time so that
            DependencyInjectors may cooperate on logic that needs to be
            duplicated, such as provisioning a session.
        """


@attr.s(auto_attribs=True)
class EarlyExit(Exception):
    """
    An L{EarlyExit} may be raised by any of the code that runs in the
    before-request dependency injection code path when using
    L{klein.Requirer.require}.

    @ivar alternateReturnValue: The return value which should instead be
        supplied as the route's response.
    """

    alternateReturnValue: "KleinRenderable"



# -*- test-case-name: klein.test.test_plating -*-

"""
Templating wrapper support for Klein.
"""
from __future__ import annotations

from functools import partial
from json import dumps
from operator import setitem
from typing import Any, Callable, Generator, List, Tuple, cast

import attr

from twisted.internet.defer import inlineCallbacks
from twisted.web.error import MissingRenderMethod
from twisted.web.iweb import IRequest
from twisted.web.template import Element, Tag, TagLoader

from ._app import _call
from ._decorators import bindable, modified, originalName


StackType = List[Tuple[Any, Callable[[Any], None]]]

# https://github.com/python/mypy/issues/224
ATOM_TYPES = (
    cast(Tuple[Any, ...], (int,))
    + cast(Tuple[Any, ...], (str,))
    + cast(Tuple[Any, ...], (float, None.__class__))
)


def _should_return_json(request: IRequest) -> bool:
    """
    Should the given request result in a JSON entity-body?
    """
    return bool(request.args.get(b"json"))


@inlineCallbacks
def resolveDeferredObjects(root: Any) -> Generator[Any, object, Any]:
    """
    Wait on possibly nested L{Deferred}s that represent a JSON
    serializable object.

    @param root: JSON-serializable object that may contain
        L{Deferred}s that resolve to JSON-serializable objects, or a
        L{Deferred} that resolves to one.

    @return: A L{Deferred} that fires with a L{Deferred}-free version
        of C{root}, or that fails with the first exception
        encountered.
    """

    result = [None]
    setResult = partial(setitem, result, 0)
    stack: StackType = [(root, setResult)]

    while stack:
        mightBeDeferred, setter = stack.pop()
        # inlineCallbacks pauses the generator only on yielded
        # Deferreds. It's resumed immediately with any other object.
        # Consequently coroutines must be wrapped in ensureDeferred.
        obj = yield mightBeDeferred
        if isinstance(obj, ATOM_TYPES):
            setter(obj)
        elif isinstance(obj, list):
            parent: Any = [None] * len(obj)
            setter(parent)
            stack.extend(
                reversed(
                    [
                        (child, partial(setitem, parent, i))
                        for i, child in enumerate(obj)
                    ]
                )
            )
        elif isinstance(obj, tuple):
            parent = [None] * len(obj)
            setter(tuple(parent))

            def setTupleItem(i, value, parent=parent, setter=setter):
                parent[i] = value
                setter(tuple(parent))

            stack.extend(
                reversed(
                    [
                        (child, partial(setTupleItem, i))
                        for i, child in enumerate(obj)
                    ]
                )
            )
        elif isinstance(obj, dict):
            parent = {}
            setter(parent)
            for key, value in reversed(list(obj.items())):
                pair = [None, None]
                setKey = partial(setitem, pair, 0)

                def setValue(value, pair=pair, parent=parent):
                    pair[1] = value
                    parent.update([pair])

                stack.append((value, setValue))
                stack.append((key, setKey))
        elif isinstance(obj, PlatedElement):
            stack.append((obj._asJSON(), setter))
        else:
            raise TypeError(
                obj,
                f"{obj} not JSON serializable",
            )

    return result[0]


def _extra_types(input):
    """
    Renderability for a few additional types.
    """
    if isinstance(input, (float, int)):
        return str(input)
    return input


class PlatedElement(Element):
    """
    The element type returned by L{Plating}.  This contains several utility
    renderers.
    """

    def __init__(
        self, slot_data, preloaded, boundInstance, presentationSlots, renderers
    ):
        """
        @param slot_data: A dictionary mapping names to values.

        @param preloaded: The pre-loaded data.
        """
        self.slot_data = slot_data
        self._boundInstance = boundInstance
        self._presentationSlots = presentationSlots
        self._renderers = renderers
        super().__init__(
            loader=TagLoader(
                preloaded.fillSlots(
                    **{k: _extra_types(v) for k, v in slot_data.items()}
                )
            )
        )

    def _asJSON(self):
        """
        Render this L{PlatedElement} as JSON-serializable data.
        """
        json_data = self.slot_data.copy()
        for ignored in self._presentationSlots:
            json_data.pop(ignored, None)
        return json_data

    def lookupRenderMethod(self, name):
        """
        @return: a renderer.
        """
        if name in self._renderers:
            wrapped = self._renderers[name]

            @modified("plated render wrapper", wrapped)
            def renderWrapper(
                request: IRequest, tag: Tag, *args: Any, **kw: Any
            ) -> Any:
                return _call(
                    self._boundInstance, wrapped, request, tag, *args, **kw
                )

            return renderWrapper
        if ":" not in name:
            raise MissingRenderMethod(self, name)
        slot, type = name.split(":", 1)

        def renderList(request, tag):
            for item in self.slot_data[slot]:
                yield tag.fillSlots(item=_extra_types(item))

        types = {
            "list": renderList,
        }
        if type in types:
            return types[type]
        else:
            raise MissingRenderMethod(self, name)


class Plating:
    """
    A L{Plating} is a container which can be used to generate HTML from data.

    Its name is derived both from tem-I{plating} and I{chrome plating}.
    """

    CONTENT = "klein:plating:content"

    def __init__(self, defaults=None, tags=None, presentation_slots=()):
        """ """
        self._defaults = {} if defaults is None else defaults
        self._loader = TagLoader(tags)
        self._presentationSlots = {self.CONTENT} | set(presentation_slots)
        self._renderers = {}

    def renderMethod(self, renderer):
        """
        Add a render method to this L{Plating} object that can be used in the
        top-level template.

        The name of the renderer to use within the template is the name of the
        decorated function.
        """
        self._renderers[str(originalName(renderer))] = renderer
        return renderer

    def routed(self, routing, tags):
        """ """

        def mydecorator(method):
            loader = TagLoader(tags)

            @modified("plating route renderer", method, routing)
            @bindable
            @inlineCallbacks
            def mymethod(
                instance: Any, request: IRequest, *args: Any, **kw: Any
            ) -> Any:
                data = yield _call(instance, method, request, *args, **kw)
                if _should_return_json(request):
                    json_data = self._defaults.copy()
                    json_data.update(data)
                    for ignored in self._presentationSlots:
                        json_data.pop(ignored, None)
                    request.setHeader(b"content-type", b"application/json")
                    ready = yield resolveDeferredObjects(json_data)
                    result = dumps(ready)
                else:
                    data[self.CONTENT] = loader.load()
                    request.setHeader(
                        b"content-type", b"text/html; charset=utf-8"
                    )
                    result = self._elementify(instance, data)
                return result

            return method

        return mydecorator

    def _elementify(self, instance, to_fill_with):
        """
        Convert this L{Plating} into a L{PlatedElement}.
        """
        slot_data = self._defaults.copy()
        slot_data.update(to_fill_with)
        [loaded] = self._loader.load()
        loaded = loaded.clone()  # type: ignore[union-attr]
        return PlatedElement(
            slot_data=slot_data,
            preloaded=loaded,
            renderers=self._renderers,
            boundInstance=instance,
            presentationSlots=self._presentationSlots,
        )

    @attr.s(auto_attribs=True)
    class _Widget:
        """
        Implementation of L{Plating.widgeted}.  This is a L{callable}
        descriptor that records the instance to which its wrapped
        function is bound, if any.  Its L{widget} method then passes
        that instance or L{None} and the result of invoking the
        function (or now bound method) to the creating L{Plating}
        instance's L{Plating._elementify} to construct a
        L{PlatedElement}.
        """

        _plating: Plating
        _function: Callable[..., Any]
        _instance: object

        def __call__(self, *args, **kwargs):
            return self._function(*args, **kwargs)

        def __get__(self, instance, owner=None):
            return self.__class__(
                self._plating,
                self._function.__get__(instance, owner),
                instance=instance,
            )

        def widget(self, *args, **kwargs):
            """
            Construct a L{PlatedElement} the rendering of this widget.
            """
            data = self._function(*args, **kwargs)
            return self._plating._elementify(self._instance, data)

        def __getattr__(self, attr):
            return getattr(self._function, attr)

    def widgeted(self, function):
        """
        A decorator that turns a function into a renderer for an
        element without a L{Klein.route}.  Use this to create reusable
        template elements.
        """
        return self._Widget(self, function, None)



import os
from io import BytesIO
from types import MappingProxyType
from typing import Any, List, Mapping, Optional, Sequence, cast
from unittest.mock import ANY, Mock, call
from urllib.parse import parse_qs

from werkzeug.exceptions import NotFound

from twisted.internet.defer import CancelledError, Deferred, fail, succeed
from twisted.internet.error import ConnectionLost
from twisted.internet.interfaces import IProducer
from twisted.internet.unix import Server
from twisted.python.failure import Failure
from twisted.trial.unittest import SynchronousTestCase
from twisted.web.http_headers import Headers
from twisted.web.iweb import IRequest
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET, Request, Site
from twisted.web.static import File
from twisted.web.template import Element, Tag, XMLString, renderer
from twisted.web.test.test_web import DummyChannel

from .. import Klein, KleinRenderable
from .._interfaces import IKleinRequest
from .._resource import (
    KleinResource,
    URLDecodeError,
    ensure_utf8_bytes,
    extractURLparts,
)
from .util import EqualityTestsMixin, recover


emptyMapping: Mapping[Any, Any] = MappingProxyType({})


class MockRequest(Request):
    finished: bool
    startedWriting: bool
    processingFailed: Mock
    setResponseCode: Mock
    setHeader: Mock

    def __init__(
        self,
        path: bytes,
        method: bytes = b"GET",
        host: bytes = b"localhost",
        port: int = 8080,
        isSecure: bool = False,
        body: bytes = b"",
        headers: Mapping[bytes, Sequence[bytes]] = emptyMapping,
    ):
        super().__init__(DummyChannel(), False)

        if not headers:
            headers = {}

        if not body:
            body = b""

        path, qpath = (path.split(b"?", 1) + [b""])[:2]

        self.site = Mock(Site)
        self.gotLength(len(body))
        self.content = BytesIO()
        self.content.write(body)
        self.content.seek(0)
        self.args = parse_qs(qpath)
        self.selfHeaders = Headers(headers)
        self.setHost(host, port, isSecure)
        self.uri = path
        self.prepath: List[bytes] = []
        self.postpath = path.split(b"/")[1:]
        self.method = method
        self.clientproto = b"HTTP/1.1"

        self.setHeader = Mock(wraps=self.setHeader)
        self.setResponseCode = Mock(wraps=self.setResponseCode)

        self._written = BytesIO()
        self.finishCount = 0
        self.writeCount = 0

        self.processingFailed = Mock(wraps=self.processingFailed)

    def registerProducer(self, producer: IProducer, streaming: bool) -> None:
        self.producer = producer
        for _ in range(2):
            if self.producer:
                # typing note: server.Request.registerProducer takes an
                # IProducer, which does not have resumeProducing.
                # This seems to expect either an IPullProducer or an
                # IPushProducer.
                self.producer.resumeProducing()  # type: ignore[attr-defined]

    def unregisterProducer(self) -> None:
        self.producer = None

    def finish(self) -> None:
        self.finishCount += 1

        if not self.startedWriting:
            self.write(b"")

        if not self.finished:
            self.finished = True
            self._cleanup()

    def write(self, data: bytes) -> None:
        self.writeCount += 1
        self.startedWriting = True

        if not self.finished:
            self._written.write(data)
        else:
            raise RuntimeError(
                "Request.write called on a request after "
                "Request.finish was called."
            )

    def getWrittenData(self) -> bytes:
        return self._written.getvalue()


def _render(
    resource: KleinResource, request: IRequest, notifyFinish: bool = True
) -> Deferred:
    result = resource.render(request)

    if isinstance(result, bytes):
        request.write(result)
        request.finish()
        return succeed(None)

    if result is not NOT_DONE_YET:  # type: ignore[comparison-overlap]
        raise AssertionError("unreachable")  # pragma: no cover

    if request.finished or not notifyFinish:  # type: ignore[attr-defined]
        return succeed(None)

    return request.notifyFinish()  # type: ignore[no-any-return,attr-defined]


class SimpleElement(Element):
    loader = XMLString(
        '<h1 xmlns:t="http://twistedmatrix.com/ns/twisted.web.template/0.1" '
        't:render="name" />'
    )

    def __init__(self, name: str) -> None:
        self._name = name

    @renderer
    def name(self, request: IRequest, tag: Tag) -> Tag:
        return tag(self._name)


class DeferredElement(SimpleElement):
    deferred: "Deferred[None]"

    @renderer
    def name(self, request: IRequest, tag: Tag) -> "Deferred[None]":
        self.deferred: "Deferred[None]" = Deferred()
        self.deferred.addCallback(lambda ignored: tag(self._name))
        return self.deferred


class LeafResource(Resource):
    isLeaf = True

    content = b"I am a leaf in the wind."

    def render(self, request: IRequest) -> bytes:
        return self.content


class ChildResource(Resource):
    isLeaf = True

    def __init__(self, name: bytes) -> None:
        self._name = name

    def render(self, request: IRequest) -> bytes:
        return b"I'm a child named " + self._name + b"!"


class ChildrenResource(Resource):
    def render(self, request: IRequest) -> bytes:
        return b"I have children!"

    def getChild(self, path: bytes, request: IRequest) -> Resource:
        if path == b"":
            return self

        return ChildResource(path)


class ProducingResource(Resource):
    def __init__(self, path: bytes, strings: List[bytes]) -> None:
        self.path = path
        self.strings = strings

    def render_GET(self, request: IRequest) -> bytes:
        producer = MockProducer(request, self.strings)
        producer.start()
        # typing note: return type should have been
        # Union[bytes, Literal[NOT_DONE_YET]] but NOT_DONE_YET is an Any
        # right now, so Literal won't accept it.
        return cast(bytes, NOT_DONE_YET)


class MockProducer:
    def __init__(self, request: IRequest, strings: List[bytes]) -> None:
        self.request = request
        self.strings = strings

    def start(self) -> None:
        self.request.registerProducer(self, False)  # type: ignore[attr-defined]

    def resumeProducing(self) -> None:
        if self.strings:
            self.request.write(self.strings.pop(0))
        else:
            self.request.unregisterProducer()  # type: ignore[attr-defined]
            self.request.finish()


class KleinResourceEqualityTests(SynchronousTestCase, EqualityTestsMixin):
    """
    Tests for L{KleinResource}'s implementation of C{==} and C{!=}.
    """

    class _One:
        oneKlein = Klein()

        @oneKlein.route("/foo")
        def foo(self, request: IRequest) -> KleinRenderable:
            pass

    _one = _One()

    class _Another:
        anotherKlein = Klein()

        @anotherKlein.route("/bar")
        def bar(self, request: IRequest) -> KleinRenderable:
            pass

    _another = _Another()

    def anInstance(self) -> KleinResource:
        return self._one.oneKlein.resource()

    def anotherInstance(self) -> KleinResource:
        return self._another.anotherKlein.resource()


class KleinResourceTests(SynchronousTestCase):
    def setUp(self) -> None:
        self.app = Klein()
        self.kr = KleinResource(self.app)

    def assertFired(self, deferred: Deferred, result: object = None) -> None:
        """
        Assert that the given deferred has fired with the given result.
        """
        self.assertEqual(self.successResultOf(deferred), result)

    def assertNotFired(self, deferred: Deferred) -> None:
        """
        Assert that the given deferred has not fired with a result.
        """
        _pawn = object()
        result = getattr(deferred, "result", _pawn)
        if result != _pawn:
            self.fail(
                "Expected deferred not to have fired, but it has: {!r}".format(
                    deferred
                )
            )

    def test_simplePost(self) -> None:
        app = self.app

        # The order in which these functions are defined
        # matters.  If the more generic one is defined first
        # then it will eat requests that should have been handled
        # by the more specific handler.

        @app.route("/", methods=["POST"])
        def handle_post(request: IRequest) -> KleinRenderable:
            return b"posted"

        @app.route("/")
        def handle_default(request: IRequest) -> KleinRenderable:
            return b"gotted"

        request = MockRequest(b"/", b"POST")
        request2 = MockRequest(b"/")

        d = _render(self.kr, request)
        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"posted")

        d2 = _render(self.kr, request2)
        self.assertFired(d2)
        self.assertEqual(request2.getWrittenData(), b"gotted")

    def test_simpleRouting(self) -> None:
        app = self.app

        @app.route("/")
        def slash(request: IRequest) -> KleinRenderable:
            return b"ok"

        request = MockRequest(b"/")

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"ok")

    def test_branchRendering(self) -> None:
        app = self.app

        @app.route("/", branch=True)
        def slash(request: IRequest) -> KleinRenderable:
            return b"ok"

        request = MockRequest(b"/foo")

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"ok")

    def test_branchWithExplicitChildrenRouting(self) -> None:
        app = self.app

        @app.route("/")
        def slash(request: IRequest) -> KleinRenderable:
            return b"ok"

        @app.route("/zeus")
        def wooo(request: IRequest) -> KleinRenderable:
            return b"zeus"

        request = MockRequest(b"/zeus")
        request2 = MockRequest(b"/")

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"zeus")

        d2 = _render(self.kr, request2)

        self.assertFired(d2)
        self.assertEqual(request2.getWrittenData(), b"ok")

    def test_branchWithExplicitChildBranch(self) -> None:
        app = self.app

        @app.route("/", branch=True)
        def slash(request: IRequest) -> KleinRenderable:
            return b"ok"

        @app.route("/zeus/", branch=True)
        def wooo(request: IRequest) -> KleinRenderable:
            return b"zeus"

        request = MockRequest(b"/zeus/foo")
        request2 = MockRequest(b"/")

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"zeus")

        d2 = _render(self.kr, request2)

        self.assertFired(d2)
        self.assertEqual(request2.getWrittenData(), b"ok")

    def test_deferredRendering(self) -> None:
        app = self.app

        deferredResponse: "Deferred[bytes]" = Deferred()

        @app.route("/deferred")
        def deferred(request: IRequest) -> KleinRenderable:
            return deferredResponse

        request = MockRequest(b"/deferred")

        d = _render(self.kr, request)

        self.assertNotFired(d)

        deferredResponse.callback(b"ok")

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"ok")

    def test_asyncRendering(self) -> None:
        app = self.app
        resource = self.kr

        request = MockRequest(b"/resource/leaf")

        @app.route("/resource/leaf")
        async def leaf(request: IRequest) -> LeafResource:
            return LeafResource()

        self.assertFired(_render(resource, request))

        self.assertEqual(request.getWrittenData(), LeafResource.content)

    def test_elementRendering(self) -> None:
        app = self.app

        @app.route("/element/<string:name>")
        def element(request: IRequest, name: str) -> KleinRenderable:
            return SimpleElement(name)

        request = MockRequest(b"/element/foo")

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(
            request.getWrittenData(), b"<!DOCTYPE html>\n<h1>foo</h1>"
        )

    def test_deferredElementRendering(self) -> None:
        app = self.app

        elements = []

        @app.route("/element/<string:name>")
        def element(request: IRequest, name: str) -> KleinRenderable:
            it = DeferredElement(name)
            elements.append(it)
            return it

        request = MockRequest(b"/element/bar")

        d = _render(self.kr, request)
        self.assertEqual(len(elements), 1)
        [oneElement] = elements
        self.assertNoResult(d)
        oneElement.deferred.callback(None)
        self.assertFired(d)
        self.assertEqual(
            request.getWrittenData(), b"<!DOCTYPE html>\n<h1>bar</h1>"
        )

    def test_leafResourceRendering(self) -> None:
        app = self.app

        request = MockRequest(b"/resource/leaf")

        @app.route("/resource/leaf")
        def leaf(request: IRequest) -> KleinRenderable:
            return LeafResource()

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), LeafResource.content)

    def test_childResourceRendering(self) -> None:
        app = self.app
        request = MockRequest(b"/resource/children/betty")

        @app.route("/resource/children/", branch=True)
        def children(request: IRequest) -> KleinRenderable:
            return ChildrenResource()

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"I'm a child named betty!")

    def test_childrenResourceRendering(self) -> None:
        app = self.app

        request = MockRequest(b"/resource/children/")

        @app.route("/resource/children/", branch=True)
        def children(request: IRequest) -> KleinRenderable:
            return ChildrenResource()

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"I have children!")

    def test_producerResourceRendering(self) -> None:
        """
        Test that Klein will correctly handle producing L{Resource}s.

        Producing Resources close the connection by themselves, sometimes after
        Klein has 'finished'. This test lets Klein finish its handling of the
        request before doing more producing.
        """
        app = self.app

        request = MockRequest(b"/resource")

        @app.route("/resource", branch=True)
        def producer(request: IRequest) -> KleinRenderable:
            return ProducingResource(request.uri, [b"a", b"b", b"c", b"d"])

        d = _render(self.kr, request, notifyFinish=False)

        self.assertNotEqual(
            request.getWrittenData(),
            b"abcd",
            "The full response should not have been written at this point.",
        )

        while request.producer:
            request.producer.resumeProducing()

        self.assertEqual(self.successResultOf(d), None)
        self.assertEqual(request.getWrittenData(), b"abcd")
        self.assertEqual(request.writeCount, 4)
        self.assertEqual(request.finishCount, 1)
        self.assertEqual(request.producer, None)

    def test_notFound(self) -> None:
        request = MockRequest(b"/fourohofour")

        d = _render(self.kr, request)

        self.assertFired(d)
        setResponseCode = request.setResponseCode
        setResponseCode.assert_called_with(404)
        self.assertIn(b"404 Not Found", request.getWrittenData())

    def test_renderUnicode(self) -> None:
        app = self.app

        request = MockRequest(b"/snowman")

        @app.route("/snowman")
        def snowman(request: IRequest) -> KleinRenderable:
            return "\u2603"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"\xE2\x98\x83")

    def test_renderNone(self) -> None:
        app = self.app

        request = MockRequest(b"/None")

        @app.route("/None")
        def none(request: IRequest) -> KleinRenderable:
            return None

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"")
        self.assertEqual(request.finishCount, 1)
        self.assertEqual(request.writeCount, 1)

    def test_staticRoot(self) -> None:
        app = self.app

        request = MockRequest(b"/__init__.py")
        with open(
            os.path.join(os.path.dirname(__file__), "__init__.py"), "rb"
        ) as f:
            expected = f.read()

        @app.route("/", branch=True)
        def root(request: IRequest) -> KleinRenderable:
            return File(os.path.dirname(__file__))

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), expected)
        self.assertEqual(request.finishCount, 1)

    def test_explicitStaticBranch(self) -> None:
        app = self.app

        request = MockRequest(b"/static/__init__.py")
        with open(
            os.path.join(os.path.dirname(__file__), "__init__.py"), "rb"
        ) as f:
            expected = f.read()

        @app.route("/static/", branch=True)
        def root(request: IRequest) -> KleinRenderable:
            return File(os.path.dirname(__file__))

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), expected)
        self.assertEqual(request.writeCount, 1)
        self.assertEqual(request.finishCount, 1)

    def test_staticDirlist(self) -> None:
        app = self.app

        request = MockRequest(b"/")

        @app.route("/", branch=True)
        def root(request: IRequest) -> KleinRenderable:
            return File(os.path.dirname(__file__))

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertIn(b"Directory listing", request.getWrittenData())
        self.assertEqual(request.writeCount, 1)
        self.assertEqual(request.finishCount, 1)

    def test_addSlash(self) -> None:
        app = self.app
        request = MockRequest(b"/foo")

        @app.route("/foo/")
        def foo(request: IRequest) -> KleinRenderable:
            return "foo"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(
            request.setHeader.call_count,
            3,
        )
        request.setHeader.assert_has_calls(
            [
                call(b"Content-Type", b"text/html; charset=utf-8"),
                call(b"Content-Length", ANY),
                call(b"Location", b"http://localhost:8080/foo/"),
            ]
        )

    def test_methodNotAllowed(self) -> None:
        app = self.app
        request = MockRequest(b"/foo", method=b"DELETE")

        @app.route("/foo", methods=["GET"])
        def foo(request: IRequest) -> KleinRenderable:
            return "foo"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.code, 405)

    def test_methodNotAllowedWithRootCollection(self) -> None:
        app = self.app
        request = MockRequest(b"/foo/bar", method=b"DELETE")

        @app.route("/foo/bar", methods=["GET"])
        def foobar(request: IRequest) -> KleinRenderable:
            return b"foo/bar"

        @app.route("/foo/", methods=["DELETE"])
        def foo(request: IRequest) -> KleinRenderable:
            return b"foo"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.code, 405)

    def test_noImplicitBranch(self) -> None:
        app = self.app
        request = MockRequest(b"/foo")

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            return b"foo"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.code, 404)

    def test_strictSlashes(self) -> None:
        app = self.app
        request = MockRequest(b"/foo/bar")

        request_url = [None]

        @app.route("/foo/bar/", strict_slashes=False)
        def root(request: IRequest) -> KleinRenderable:
            request_url[0] = request.URLPath()
            return b"foo"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(str(request_url[0]), "http://localhost:8080/foo/bar")
        self.assertEqual(request.getWrittenData(), b"foo")
        self.assertEqual(request.code, 200)

    def test_URLPath(self) -> None:
        app = self.app
        request = MockRequest(b"/egg/chicken")

        request_url = [None]

        @app.route("/egg/chicken")
        def wooo(request: IRequest) -> KleinRenderable:
            request_url[0] = request.URLPath()
            return b"foo"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(
            str(request_url[0]), "http://localhost:8080/egg/chicken"
        )

    def test_URLPath_root(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        request_url = [None]

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            request_url[0] = request.URLPath()
            return b""

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(str(request_url[0]), "http://localhost:8080/")

    def test_URLPath_traversedResource(self) -> None:
        app = self.app
        request = MockRequest(b"/resource/foo")

        request_url = [None]

        class URLPathResource(Resource):
            def render(self, request: IRequest) -> KleinRenderable:
                request_url[0] = request.URLPath()
                return b""

            def getChild(  # type: ignore[override]
                self,
                request: IRequest,
                path: bytes,
            ) -> Resource:
                return self

        @app.route("/resource/", branch=True)
        def root(request: IRequest) -> KleinRenderable:
            return URLPathResource()

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(
            str(request_url[0]), "http://localhost:8080/resource/foo"
        )

    def test_handlerRaises(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        failures = []

        class RouteFailureTest(Exception):
            pass

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            def _capture_failure(f: Failure) -> Failure:
                failures.append(f)
                return f

            return fail(RouteFailureTest("die")).addErrback(_capture_failure)

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.code, 500)
        processingFailed = request.processingFailed
        processingFailed.assert_called_once_with(failures[0])
        self.flushLoggedErrors(RouteFailureTest)

    def test_genericErrorHandler(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        failures = []

        class RouteFailureTest(Exception):
            pass

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            raise RouteFailureTest("not implemented")

        @app.handle_errors
        def handle_errors(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            failures.append(failure)
            request.setResponseCode(501)
            return b""

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.code, 501)
        assert not request.processingFailed.called

    def test_typeSpecificErrorHandlers(self) -> None:
        app = self.app
        request = MockRequest(b"/")
        type_error_handled = [False]
        generic_error_handled = [False]

        failures = []

        class TypeFilterTestError(Exception):
            pass

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            return fail(TypeFilterTestError("not implemented"))

        @app.handle_errors(TypeError)
        def handle_type_error(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            type_error_handled[0] = True
            return b""

        @app.handle_errors(TypeFilterTestError)
        def handle_type_filter_test_error(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            failures.append(failure)
            request.setResponseCode(501)
            return b""

        @app.handle_errors
        def handle_generic_error(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            generic_error_handled[0] = True
            return b""

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(
            request.processingFailed.called,
            False,
        )
        self.assertEqual(type_error_handled[0], False)
        self.assertEqual(generic_error_handled[0], False)
        self.assertEqual(len(failures), 1)
        self.assertEqual(request.code, 501)

        # Test the above handlers, which otherwise lack test coverage.

        @app.route("/type_error")
        def type_error(request: IRequest) -> KleinRenderable:
            return fail(TypeError("type error"))

        d = _render(self.kr, MockRequest(b"/type_error"))
        self.assertFired(d)
        self.assertEqual(type_error_handled[0], True)

        @app.route("/generic_error")
        def generic_error(request: IRequest) -> KleinRenderable:
            return fail(Exception("generic error"))

        d = _render(self.kr, MockRequest(b"/generic_error"))
        self.assertFired(d)
        self.assertEqual(generic_error_handled[0], True)

    def test_notFoundException(self) -> None:
        app = self.app
        request = MockRequest(b"/")
        generic_error_handled = [False]

        @app.handle_errors(NotFound)
        def handle_not_found(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            request.setResponseCode(404)
            return b"Custom Not Found"

        @app.handle_errors
        def handle_generic_error(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            generic_error_handled[0] = True
            return b""

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(
            request.processingFailed.called,
            False,
        )
        self.assertEqual(generic_error_handled[0], False)
        self.assertEqual(request.code, 404)
        self.assertEqual(request.getWrittenData(), b"Custom Not Found")
        self.assertEqual(request.writeCount, 1)

        # Test the above handlers, which otherwise lack test coverage.

        @app.route("/generic_error")
        def generic_error(request: IRequest) -> KleinRenderable:
            return fail(Exception("generic error"))

        d = _render(self.kr, MockRequest(b"/generic_error"))
        self.assertFired(d)
        self.assertEqual(generic_error_handled[0], True)

    def test_errorHandlerNeedsRendering(self) -> None:
        """
        Renderables returned by L{handle_errors} are rendered.
        """
        app = self.app
        request = MockRequest(b"/")

        @app.handle_errors(NotFound)
        def handle_not_found(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            return SimpleElement("Not Found Element")

        d = _render(self.kr, request)

        rendered = b"<!DOCTYPE html>\n<h1>Not Found Element</h1>"

        self.assertFired(d)
        self.assertEqual(
            request.processingFailed.called,
            False,
        )
        self.assertEqual(request.getWrittenData(), rendered)

    def test_errorHandlerReturnsResource(self) -> None:
        """
        Resources returned by L{Klein.handle_errors} are rendered
        """
        app = self.app
        request = MockRequest(b"/")

        class NotFoundResource(Resource):
            isLeaf = True

            def render(self, request: IRequest) -> KleinRenderable:
                request.setResponseCode(404)
                return b"Nothing found"

        @app.handle_errors(NotFound)
        def handle_not_found(
            request: IRequest, failure: Failure
        ) -> KleinRenderable:
            return NotFoundResource()

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.code, 404)
        self.assertEqual(request.getWrittenData(), b"Nothing found")

    def test_requestWriteAfterFinish(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            request.finish()
            return b"foo"

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.writeCount, 2)
        self.assertEqual(request.getWrittenData(), b"")
        [failure] = self.flushLoggedErrors(RuntimeError)

        self.assertEqual(
            str(failure.value),
            (
                "Request.write called on a request after Request.finish was "
                "called."
            ),
        )

    def test_requestFinishAfterConnectionLost(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        finished: "Deferred[bytes]" = Deferred()

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            assert isinstance(request, Request)
            request.notifyFinish().addBoth(lambda _: finished.callback(b"foo"))
            return finished

        d = _render(self.kr, request)

        def _eb(result: object) -> None:
            [failure] = self.flushLoggedErrors(RuntimeError)

            self.assertEqual(
                str(failure.value),
                (
                    "Request.finish called on a request after its connection "
                    "was lost; use Request.notifyFinish to keep track of this."
                ),
            )

        d.addErrback(lambda _: finished)
        d.addErrback(_eb)

        self.assertNotFired(d)

        request.connectionLost(ConnectionLost())

        self.assertFired(d)

    def test_routeHandlesRequestFinished(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        cancelled: List[Failure] = []

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            assert isinstance(request, Request)
            _d: "Deferred[bytes]" = Deferred()
            _d.addErrback(cancelled.append)
            request.notifyFinish().addCallback(lambda _: _d.cancel())
            return _d

        d = _render(self.kr, request)

        request.finish()

        self.assertFired(d)

        cancelled[0].trap(CancelledError)
        self.assertEqual(request.getWrittenData(), b"")
        self.assertEqual(request.writeCount, 1)
        self.assertEqual(
            request.processingFailed.call_count,
            0,
        )

    def test_url_for(self) -> None:
        app = self.app
        request = MockRequest(b"/foo/1")

        relative_url: List[str] = ["** ROUTE NOT CALLED **"]

        @app.route("/foo/<int:bar>")
        def foo(request: IRequest, bar: int) -> KleinRenderable:
            krequest = IKleinRequest(request)
            relative_url[0] = krequest.url_for("foo", {"bar": bar + 1})
            return b""

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(relative_url[0], "/foo/2")

    def test_cancelledDeferred(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        inner_d: "Deferred[bytes]" = Deferred()

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            return inner_d

        d = _render(self.kr, request)

        inner_d.cancel()

        self.assertFired(d)
        self.flushLoggedErrors(CancelledError)

    def test_external_url_for(self) -> None:
        app = self.app
        request = MockRequest(b"/foo/1")

        relative_url: List[Optional[str]] = [None]

        @app.route("/foo/<int:bar>")
        def foo(request: IRequest, bar: int) -> KleinRenderable:
            krequest = IKleinRequest(request)
            relative_url[0] = krequest.url_for(
                "foo", {"bar": bar + 1}, force_external=True
            )
            return b""

        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(relative_url[0], "http://localhost:8080/foo/2")

    def test_cancelledIsEatenOnConnectionLost(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            _d: "Deferred[bytes]" = Deferred()
            assert isinstance(request, Request)
            request.notifyFinish().addErrback(lambda _: _d.cancel())
            return _d

        d = _render(self.kr, request)

        self.assertNotFired(d)

        request.connectionLost(ConnectionLost())

        def _cb(result: object) -> None:
            processingFailed = request.processingFailed
            self.assertEqual(
                processingFailed.call_count,
                0,
            )

        d = recover(d, ConnectionLost)
        d.addCallback(_cb)
        self.assertFired(d)

    def test_cancelsOnConnectionLost(self) -> None:
        app = self.app
        request = MockRequest(b"/")

        handler_d: "Deferred[bytes]" = Deferred()

        @app.route("/")
        def root(request: IRequest) -> KleinRenderable:
            return handler_d

        d = _render(self.kr, request)

        self.assertNotFired(d)

        request.connectionLost(ConnectionLost())

        handler_d = recover(handler_d, CancelledError)

        d = recover(d, ConnectionLost)
        d.addCallback(lambda _: handler_d)
        self.assertFired(d)

    def test_ensure_utf8_bytes(self) -> None:
        self.assertEqual(ensure_utf8_bytes("abc"), b"abc")
        self.assertEqual(ensure_utf8_bytes("\u2202"), b"\xe2\x88\x82")
        self.assertEqual(ensure_utf8_bytes(b"\xe2\x88\x82"), b"\xe2\x88\x82")

    def test_decodesPath(self) -> None:
        """
        server_name, path_info, and script_name are decoded as UTF-8 before
        being handed to werkzeug.
        """
        request = MockRequest(b"/f\xc3\xb6\xc3\xb6")

        _render(self.kr, request)
        kreq = IKleinRequest(request)
        self.assertIsInstance(kreq.mapper.server_name, str)
        self.assertIsInstance(kreq.mapper.path_info, str)
        self.assertIsInstance(kreq.mapper.script_name, str)

    def test_failedDecodePathInfo(self) -> None:
        """
        If decoding of one of the URL parts (in this case PATH_INFO) fails, the
        error is logged and 400 returned.
        """
        request = MockRequest(b"/f\xc3\xc3\xb6")
        _render(self.kr, request)
        rv = request.getWrittenData()
        self.assertEqual(b"Non-UTF-8 encoding in URL.", rv)
        self.assertEqual(1, len(self.flushLoggedErrors(UnicodeDecodeError)))

    def test_urlDecodeErrorRepr(self) -> None:
        """
        URLDecodeError.__repr__ formats properly.
        """
        try:
            raise ValueError("value")
        except ValueError:
            valueFailure = Failure()

        try:
            raise TypeError("type")
        except TypeError:
            typeFailure = Failure()

        error = URLDecodeError([("VALUE", valueFailure), ("TYPE", typeFailure)])
        self.assertEqual(
            "<URLDecodeError(errors=["
            "('VALUE', "
            "<twisted.python.failure.Failure builtins.ValueError: value>), "
            "('TYPE', "
            "<twisted.python.failure.Failure builtins.TypeError: type>)])>",
            repr(error),
        )

    def test_subroutedBranch(self) -> None:
        subapp = Klein()

        @subapp.route("/foo")
        def foo(request: IRequest) -> KleinRenderable:
            return b"foo"

        app = self.app
        with app.subroute("/sub") as app:

            @app.route("/app", branch=True)
            def subapp_endpoint(request: IRequest) -> KleinRenderable:
                return subapp.resource()

        request = MockRequest(b"/sub/app/foo")
        d = _render(self.kr, request)

        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"foo")

    def test_correctContentLengthForRequestRedirect(self) -> None:
        app = self.app

        @app.route("/alias", alias=True)
        @app.route("/real")
        def real(request: IRequest) -> KleinRenderable:
            return b"42"

        request = MockRequest(b"/real")
        d = _render(self.kr, request)
        self.assertFired(d)
        self.assertEqual(request.getWrittenData(), b"42")

        request = MockRequest(b"/alias")
        d = _render(self.kr, request)
        self.assertFired(d)
        # Werkzeug switched the redirect status code used from 301 to 308.
        # Both are valid here.
        self.assertIn(
            request.setResponseCode.call_args[0],
            [(301,), (308,)],
        )

        actual_length = len(request.getWrittenData())
        values = request.responseHeaders.getRawHeaders(b"content-length")
        assert values is not None
        reported_length = int(values[0])
        self.assertEqual(reported_length, actual_length)


class ExtractURLpartsTests(SynchronousTestCase):
    """
    Tests for L{klein.resource.extractURLparts}.
    """

    def test_types(self) -> None:
        """
        Returns the correct types.
        """
        (
            url_scheme,
            server_name,
            server_port,
            path_info,
            script_name,
        ) = extractURLparts(MockRequest(b"/f\xc3\xb6\xc3\xb6"))

        self.assertIsInstance(url_scheme, str)
        self.assertIsInstance(server_name, str)
        self.assertIsInstance(server_port, int)
        self.assertIsInstance(path_info, str)
        self.assertIsInstance(script_name, str)

    def assertDecodingFailure(
        self, exception: URLDecodeError, part: str
    ) -> None:
        """
        Checks whether C{exception} consists of a single L{UnicodeDecodeError}
        for C{part}.
        """
        self.assertEqual(1, len(exception.errors))
        actualPart, actualFail = exception.errors[0]
        self.assertEqual(part, actualPart)
        self.assertIsInstance(actualFail.value, UnicodeDecodeError)

    def test_failServerName(self) -> None:
        """
        Raises URLDecodeError if SERVER_NAME can't be decoded.
        """
        request = MockRequest(b"/foo")
        request.getRequestHostname = (  # type:ignore[method-assign]
            lambda: b"f\xc3\xc3\xb6"
        )
        e = self.assertRaises(URLDecodeError, extractURLparts, request)
        self.assertDecodingFailure(e, "SERVER_NAME")

    def test_failPathInfo(self) -> None:
        """
        Raises URLDecodeError if PATH_INFO can't be decoded.
        """
        request = MockRequest(b"/f\xc3\xc3\xb6")
        e = self.assertRaises(URLDecodeError, extractURLparts, request)
        self.assertDecodingFailure(e, "PATH_INFO")

    def test_failScriptName(self) -> None:
        """
        Raises URLDecodeError if SCRIPT_NAME can't be decoded.
        """
        request = MockRequest(b"/foo")
        request.prepath = [b"f\xc3\xc3\xb6"]
        e = self.assertRaises(URLDecodeError, extractURLparts, request)
        self.assertDecodingFailure(e, "SCRIPT_NAME")

    def test_failAll(self) -> None:
        """
        If multiple parts fail, they all get appended to the errors list of
        URLDecodeError.
        """
        request = MockRequest(b"/f\xc3\xc3\xb6")
        request.prepath = [b"f\xc3\xc3\xb6"]
        request.getRequestHostname = (  # type:ignore[method-assign]
            lambda: b"f\xc3\xc3\xb6"
        )
        e = self.assertRaises(URLDecodeError, extractURLparts, request)
        self.assertEqual(
            {"SERVER_NAME", "PATH_INFO", "SCRIPT_NAME"},
            {part for part, _ in e.errors},
        )

    def test_afUnixSocket(self) -> None:
        """
        Test proper handling of AF_UNIX sockets
        """
        request = MockRequest(b"/f\xc3\xb6\xc3\xb6")
        server_mock = Mock(Server)
        server_mock.getRequestHostname = "/var/run/twisted.socket"
        request.host = server_mock
        (
            url_scheme,
            server_name,
            server_port,
            path_info,
            script_name,
        ) = extractURLparts(request)

        self.assertIsInstance(url_scheme, str)
        self.assertIsInstance(server_name, str)
        self.assertIsInstance(server_port, int)
        self.assertIsInstance(path_info, str)
        self.assertIsInstance(script_name, str)


class GlobalAppTests(SynchronousTestCase):
    """
    Tests for the global app object
    """

    def test_global_app(self) -> None:
        from klein.app import handle_errors, resource, route, run

        globalApp = run.__self__  # type: ignore[attr-defined]

        self.assertIs(
            route.__self__,  # type: ignore[attr-defined]
            globalApp,
        )
        self.assertIs(
            resource.__self__,  # type: ignore[attr-defined]
            globalApp,
        )
        self.assertIs(
            handle_errors.__self__,  # type: ignore[attr-defined]
            globalApp,
        )

        @route("/")
        def index(request: IRequest) -> KleinRenderable:
            raise RuntimeError("oops")

        @handle_errors(RuntimeError)
        def on_zero(request: IRequest, failure: Failure) -> KleinRenderable:
            return b"alive"

        request = MockRequest(b"/")
        d = _render(resource(), request)
        self.assertIsNone(self.successResultOf(d))
        self.assertEqual(request.getWrittenData(), b"alive")

    def test_weird_resource_situation(self) -> None:
        """
        Historically, the object named "C{klein.resource}" has had two
        meanings:

            - One is "C{klein.*} is sort of like a C{klein.Klein} instance, so
              C{klein.resource()} is sort of like C{klein.Klein.resource()}".

            - The other is "the public module in which
              C{klein.resource.KleinResource} is defined".

        This used to only work by accident; these meanings both sort of worked
        but only as long as you followed a certain import convention (C{from
        klein import resource} for the former, C{from klein.resource import
        KleinResource} for the latter).  This test ensures that
        C{klein.resource} is a special object, callable as you would expect
        from the former, but also having the attributes of the latter.
        """
        from klein import resource
        from klein.resource import KleinResource, ensure_utf8_bytes

        self.assertEqual(
            repr(resource), "<special bound method/module klein.resource>"
        )
        self.assertIdentical(resource.KleinResource, KleinResource)
        self.assertIdentical(resource.ensure_utf8_bytes, ensure_utf8_bytes)



from typing import Iterator, Sequence, Tuple, cast

from hyperlink import DecodedURL
from treq.testing import StubTreq
from zope.interface import Interface

from twisted.python.components import Componentized
from twisted.trial.unittest import SynchronousTestCase
from twisted.web.http_headers import Headers
from twisted.web.iweb import IRequest

from klein import Klein, RequestComponent, RequestURL, Requirer, Response
from klein.interfaces import IRequiredParameter


class BadlyBehavedHeaders(Headers):
    """
    Make L{Headers} lie, and refuse to return a Host header from
    getAllRequestHeaders.
    """

    def getAllRawHeaders(self) -> Iterator[Tuple[bytes, Sequence[bytes]]]:
        """
        Don't return a host header.
        """
        for key, values in super().getAllRawHeaders():
            if key != b"Host":
                yield (key, values)


router = Klein()
requirer = Requirer()


@requirer.require(
    router.route("/hello/world", methods=["GET"]),
    # typing note: https://github.com/Shoobx/mypy-zope/issues/39
    url=cast(IRequiredParameter, RequestURL()),
)
def requiresURL(url: DecodedURL) -> str:
    """
    This is a route that requires a URL.
    """
    text: str = url.child("hello/ world").asText()
    return text


class ISample(Interface):
    """
    Interface for testing.
    """


@requirer.prerequisite([ISample])
def provideSample(request: IRequest) -> None:
    """
    This requirer prerequisite installs a string as the provider of ISample.
    """
    cast(Componentized, request).setComponent(ISample, "sample component")


@requirer.require(
    router.route("/retrieve/component", methods=["GET"]),
    component=RequestComponent(ISample),
)
def needsComponent(component: str) -> str:
    """
    This route requires and returns an L{ISample}.
    """
    return component


@requirer.require(router.route("/set/headers"))
def someHeaders() -> Response:
    """
    Set some response attributes.
    """
    return Response(
        209,
        {"x-single-header": b"one", "x-multi-header": [b"two", b"three"]},
        "this is the response body",
    )


class RequireURLTests(SynchronousTestCase):
    """
    Tests for RequestURL() required parameter.
    """

    def test_requiresURL(self) -> None:
        """
        When RequestURL is specified to a requirer, a DecodedURL object will be
        passed in to the decorated route.
        """
        response = self.successResultOf(
            self.successResultOf(
                StubTreq(router.resource()).get(
                    "https://example.com/hello/world"
                )
            ).text()
        )

        self.assertEqual(
            response, "https://example.com/hello/world/hello%2F%20world"
        )

    def test_requiresURLNonStandardPort(self) -> None:
        """
        When RequestURL is specified to a requirer, a DecodedURL object will be
        passed in to the decorated route.
        """
        response = self.successResultOf(
            self.successResultOf(
                StubTreq(router.resource()).get(
                    "http://example.com:8080/hello/world"
                )
            ).text()
        )

        self.assertEqual(
            response, "http://example.com:8080/hello/world/hello%2F%20world"
        )

    def test_requiresURLBadlyBehavedClient(self) -> None:
        """
        requiresURL will press on in the face of badly-behaved client code.
        """
        response = self.successResultOf(
            self.successResultOf(
                StubTreq(router.resource()).get(
                    "https://example.com/hello/world",
                    headers=BadlyBehavedHeaders(),
                )
            ).text()
        )
        self.assertEqual(
            response,
            # Static values from StubTreq - this should probably be tunable.
            "https://127.0.0.1:31337/hello/world/hello%2F%20world",
        )


class RequireComponentTests(SynchronousTestCase):
    """
    Tests for RequestComponent.
    """

    def test_requestComponent(self) -> None:
        """
        Test for requiring a component installed on the request.
        """
        response = self.successResultOf(
            self.successResultOf(
                StubTreq(router.resource()).get(
                    "https://example.com/retrieve/component",
                )
            ).text()
        )
        self.assertEqual(response, "sample component")


class ResponseTests(SynchronousTestCase):
    """
    Tests for L{klein.Response}.
    """

    def test_basicResponse(self) -> None:
        """
        Since methods decorated with C{@require} don't receive the request and
        can't access it to set headers and response codes, instead, they can
        return a Response object that has those attributes.
        """
        response = self.successResultOf(
            StubTreq(router.resource()).get(
                "https://example.com/set/headers",
            )
        )
        self.assertEqual(response.code, 209)
        self.assertEqual(
            response.headers.getRawHeaders(b"X-Single-Header"), [b"one"]
        )
        self.assertEqual(
            response.headers.getRawHeaders(b"X-Multi-Header"),
            [b"two", b"three"],
        )
